
import pytest

from ctc import evm
from ctc.spec.typedefs import network_types
from ctc.config import config_defaults


@pytest.mark.parametrize(
    'test',
    [
        [1, 'ethereum'],
        ['ethereum', 'ethereum'],
        [3, 'ropsten'],
    ],
)
def test_get_network_name(test):
    network_reference, network_name = test

    actual_network_name = evm.get_network_name(network=network_reference)
    assert actual_network_name == network_name


@pytest.mark.parametrize(
    'test',
    [
        [1, 1],
        ['ethereum', 1],
        ['ropsten', 3],
    ],
)
def test_get_chain_id(test):
    network_reference, network_chain_id = test

    actual_network_chain_id = evm.get_network_chain_id(
        network=network_reference
    )
    assert actual_network_chain_id == network_chain_id


def test_get_network_metadata():
    metadata = evm.get_network_metadata(network=1)
    assert metadata['name'] == 'ethereum'
    assert metadata['chain_id'] == 1
    assert metadata['block_explorer'] == 'etherscan.io'


def test_get_networks():
    config_defaults.get_default_networks_metadata()




from __future__ import annotations

import typing

from ctc import spec
from .. import binary_utils


def abi_decode(
    data: spec.GenericBinaryData,
    types: spec.ABIDatatypeStr | typing.Sequence[spec.ABIDatatypeStr],
    *,
    convert_invalid_str_to_none: bool = False,
    convert_invalid_str_to: str | None = None,
) -> typing.Any:
    """decode ABI-encoded data, similar to solidity's abi.decode()"""

    import eth_abi_lite

    data = binary_utils.to_binary(data)

    # TODO: more general left padding for irregular sized data
    if data == b'':
        data = b'\x00' * 32

    # hardcode some conversions for speed
    if types == 'address':
        return '0x' + data[-20:].hex()
    elif types in ['uint256']:
        return int.from_bytes(data, 'big', signed=False)
    # elif types == '(uint256)':
    #     return (int.from_bytes(data, 'big', signed=False),)

    # decode data
    if isinstance(types, str):
        try:
            return eth_abi_lite.decode_single(types, data)
        except eth_abi_lite.exceptions.DecodingError as e:
            # handle improperly encoded data
            if types == 'string':
                if convert_invalid_str_to_none:
                    return None
                elif convert_invalid_str_to is not None:
                    return convert_invalid_str_to
            raise e
    else:
        if len(data) < 64:
            data = b'\x00' * (64 - len(data)) + data
        return eth_abi_lite.decode_abi(types, data)


def abi_encode(
    data: typing.Any,
    types: spec.ABIDatatypeStr | typing.Sequence[spec.ABIDatatypeStr],
) -> bytes:
    """encode data in ABI format, similar to solidity's abi.encode()"""

    import eth_abi_lite

    if isinstance(types, str):
        return eth_abi_lite.encode_single(types, data)
    else:
        return eth_abi_lite.encode_abi(types, data)


def abi_encode_packed(
    data: typing.Any,
    types: spec.ABIDatatypeStr | typing.Sequence[spec.ABIDatatypeStr],
) -> bytes:
    """encode data in ABI format, similar to solidity's abi.encode_packed()"""

    from eth_abi_lite import packed

    if isinstance(types, str):
        return packed.encode_single_packed(types, data)
    else:
        return packed.encode_abi_packed(types, data)



from __future__ import annotations

import typing

from ctc import spec
from .. import binary_utils


def get_transaction_type(
    transaction_or_type: spec.Data | typing.Mapping[str, typing.Any]
) -> int:
    """get transaction type"""

    if isinstance(transaction_or_type, (int, str, bytes)):
        # if type provided directly, convert to int
        return binary_utils.binary_convert(transaction_or_type, 'integer')

    elif isinstance(transaction_or_type, dict):
        transaction = transaction_or_type

        # if type explicitly provided, use it
        if 'type' in transaction:
            return binary_utils.binary_convert(transaction['type'], 'integer')

        # otherwise, infer type based on fields
        if 'max_fee_per_gas' in transaction:
            return 2
        if 'access_list' in transaction or 'accessList' in transaction:
            return 1
        else:
            return 0

    else:
        raise Exception('unknown transaction format')


def get_transaction_type_name(
    transaction_or_type: typing.Mapping[str, typing.Any] | spec.Data
) -> str:
    """get transaction type name"""

    transaction_type = get_transaction_type(transaction_or_type)

    # return name
    if transaction_type == 0:
        return 'legacy'
    elif transaction_type == 1:
        return 'eip2930'
    elif transaction_type == 2:
        return 'eip1559'
    else:
        raise Exception('unknown transaction type: ' + str(transaction_type))


def get_transaction_type_keys(
    transaction: typing.Mapping[str, typing.Any],
    *,
    signed: bool,
) -> tuple[str, ...]:
    """return list of keys that are used by a given transaction type"""

    transaction_type = get_transaction_type(transaction)

    # determine whether camelcase
    if transaction_type in [1, 2]:
        if 'accessList' in transaction:
            snake = False
        elif 'access_list' in transaction:
            snake = True
        else:
            raise Exception('transaction does not specify access_list or accessList')
    elif transaction_type == 0:
        snake = 'chain_id' in transaction
    else:
        raise Exception('unknown transaction type: ' + str(transaction_type))

    # get keys
    if transaction_type == 0:
        keys: tuple[str, ...] = spec.transaction_keys_legacy
    elif transaction_type == 1 and snake:
        keys = spec.transaction_keys_eip2930
    elif transaction_type == 1 and not snake:
        keys = spec.transaction_keys_eip2930_camel
    elif transaction_type == 2 and snake:
        keys = spec.transaction_keys_eip1559
    elif transaction_type == 2 and not snake:
        keys = spec.transaction_keys_eip1559_camel
    else:
        raise Exception('unknown transaction type: ' + str(transaction_type))

    # remove signature keys if unsigned
    if not signed:
        keys = tuple(key for key in keys if key not in ('v', 'r', 's'))

    return keys



from __future__ import annotations

from ... import query_utils
from . import contract_abis_statements


async_query_contract_abi = query_utils.wrap_selector_with_connection(
    contract_abis_statements.async_select_contract_abi,
    'contract_abis',
)

async_query_contract_abis = query_utils.wrap_selector_with_connection(
    contract_abis_statements.async_select_contract_abis,
    'contract_abis',
)



from __future__ import annotations

from ctc import spec

from .. import rpc_constructors
from .. import rpc_request
from .. import rpc_digestors


async def async_eth_get_compilers(
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_get_compilers()
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_eth_get_compilers(response=response)


async def async_eth_compile_lll(
    code: str, *, context: spec.Context = None
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_compile_lll(code=code)
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_eth_compile_lll(response=response)


async def async_eth_compile_solidity(
    code: str, *, context: spec.Context = None
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_compile_solidity(code=code)
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_eth_compile_solidity(response=response)


async def async_eth_compile_serpent(
    code: str, *, context: spec.Context = None
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_compile_serpent(code=code)
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_eth_compile_serpent(response=response)




"""
- revalidation after a reorg is expensive
- should use a large enough confirmation time to make reorgs rare

References on common confirmation thresholds
- https://support.nexo.io/s/article/processing-time-of-cryptocurrency-deposits-blockchain-confirmations
- https://developers.circle.com/developer/docs/confirmations
- https://support.kraken.com/hc/en-us/articles/203325283-Cryptocurrency-deposit-processing-times
"""

from __future__ import annotations

from ctc import spec


def get_required_confirmations(context: spec.Context) -> int:
    # for now use single confirmation age for all datatypes and networks
    return 128




import toolsql

from ctc import db
from ctc.protocols.chainlink_utils import chainlink_db

import conftest


async def test_get_chainlink_feed_payload():
    await chainlink_db.async_get_complete_feed_payload()


async def test_populate_feeds():
    db_config = conftest.get_test_db_config()
    with toolsql.connect(db_config) as conn:
        db.initialize_schema(
            schema_name='chainlink',
            context=dict(network='ethereum'),
            conn=conn,
        )
    async with toolsql.async_connect(db_config) as conn:
        await chainlink_db.async_import_network_to_db(
            network='ethereum',
            conn=conn,
        )


example_data = [
    {
        'address': '0x31e0a88fecb6ec0a411dbe0e9e76391498296ee9',
        'name': 'FEI / USD',
        'deviation': '1',
        'heartbeat': '1h',
        'decimals': 8,
        'asset': 'FEI Protocol',
        'asset_type': 'Crypto',
        'status': 'verified',
    },
    {
        'address': '0x5f4ec3df9cbd43714fe2740f5e3616155c5b8419',
        'name': 'ETH / USD',
        'deviation': '0.5',
        'heartbeat': '1h',
        'decimals': 8,
        'asset': 'Ethereum',
        'asset_type': 'Crypto',
        'status': 'verified',
    },
]


async def test_chainlink_crud():

    db_config = conftest.get_test_db_config()
    db_schema = db.get_prepared_schema(
        schema_name='chainlink',
        context=dict(network='ethereum'),
    )
    toolsql.create_db(
        db_config=db_config,
        db_schema=db_schema,
        if_not_exists=True,
        confirm=True,
    )

    network = 1

    # insert data
    async with toolsql.async_connect(db_config) as conn:
        for feed_data in example_data:
            await chainlink_db.async_upsert_feed(
                conn=conn,
                feed=feed_data,
                context=dict(network=network),
            )

    # get data individually
    async with toolsql.async_connect(db_config) as conn:
        for feed_data in example_data:
            db_feed = await chainlink_db.async_select_feed(
                conn=conn,
                address=feed_data['address'],
                context=dict(network=network),
            )
            for key, target_value in feed_data.items():
                assert target_value == db_feed[key]

    # get data collectively
    async with toolsql.async_connect(db_config) as conn:
        addresses = [feed_data['address'] for feed_data in example_data]
        db_feeds = await chainlink_db.async_select_feeds(
            conn=conn,
            addresses=addresses,
            context=dict(network=network),
        )
        db_feeds = sorted(db_feeds, key=lambda feed: feed['address'])
        assert db_feeds == example_data

    # delete entries one by one
    async with toolsql.async_connect(db_config) as conn:
        for feed_data in example_data:
            await chainlink_db.async_delete_feed(
                conn=conn,
                address=feed_data['address'],
                context=dict(network=network),
            )

    # ensure all entries deleted
    async with toolsql.async_connect(db_config) as conn:
        db_feeds = await chainlink_db.async_select_feeds(
            conn=conn,
            addresses=addresses,
            context=dict(network=network),
        )
        assert all(item is None for item in db_feeds)




from __future__ import annotations

from ctc import config
from ctc import spec


async def async_get_eth_total_supply(
    *,
    context: spec.Context = None,
) -> int | float:
    import aiohttp

    chain_id = config.get_context_chain_id(context)
    if chain_id != 1:
        raise NotImplementedError('total supply only implemented for ethereum')

    url = 'http://api.etherscan.io/api?module=stats&action=ethsupply'
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            if response.status != 200:
                raise Exception('could not get result')
            result = await response.json()
            return int(result['result'])



from __future__ import annotations

import functools
import math

import rich.console
import toolcli
import toolstr
import tooltime

from ctc import cli
from ctc import evm
from ctc import spec
from ctc.defi.metric_utils import ohlc_utils
from ctc.protocols import uniswap_v2_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_chart_command,
        'help': 'chart price action of uniswap pools',
        'args': [
            {
                'name': 'pool',
                'help': 'Uniswap pool address',
            },
            {
                'name': '--timescale',
                'help': 'size of candlesticks, e.g. 1h, 1d, or 1w',
            },
            {
                'name': '--invert',
                'action': 'store_true',
                'help': 'use inverse of price',
            },
            {
                'name': '--no-volume',
                'action': 'store_true',
                'help': 'hide volume data',
            },
        ],
        'examples': ['0x9928e4046d7c6513326ccea028cd3e7a91c7590a'],
    }


async def async_chart_command(
    *,
    pool: spec.Address,
    invert: bool,
    timescale: str,
    no_volume: bool,
) -> None:
    import asyncio

    metadata_task = asyncio.create_task(
        uniswap_v2_utils.async_get_pool_tokens_metadata(pool)
    )

    columns = toolcli.get_n_terminal_cols()
    n_candles = math.floor((columns - 10) / 2)
    if timescale is None:
        candle_timescale = '1d'
    else:
        candle_timescale = timescale
    candle_seconds = tooltime.timelength_to_seconds(candle_timescale)
    window_seconds = candle_seconds * n_candles

    window_end = tooltime.create_timestamp_seconds()
    window_start = (
        math.floor((window_end - window_seconds) / candle_seconds)
        * candle_seconds
    )
    start_block = await evm.async_get_block_of_timestamp(window_start) - 1
    end_block = await evm.async_get_latest_block_number()

    # get data
    swaps = await uniswap_v2_utils.async_get_pool_swaps(
        pool,
        start_block=start_block,
        end_block=end_block,
        normalize=True,
        include_volumes=True,
        include_prices=True,
    )

    # compute candlesticks
    prices = swaps['price__0__per__1'].to_numpy()
    x_volumes = swaps['volume__0'].to_numpy()
    if invert:
        prices = 1 / prices
    block_timestamps = await evm.async_get_block_timestamps(
        swaps['block_number'].to_list()
    )
    ohlc = ohlc_utils.compute_ohlc(
        values=list(prices),
        indices=block_timestamps,
        bin_size=candle_seconds,
        volumes=list(x_volumes),
    )
    ohlc = ohlc[-n_candles:]

    min_price = min(prices)
    max_price = max(prices)
    min_time = ohlc['bin'][0]
    max_time = ohlc['bin'][-1] + candle_seconds
    render_grid = toolstr.create_grid(
        n_rows=20,
        n_columns=n_candles * 2,
        xmin=min_time - 0.05 * (max_time - min_time),
        xmax=max_time + 0.05 * (max_time - min_time),
        ymin=min_price - 0.05 * (max_price - min_price),
        ymax=max_price + 0.05 * (max_price - min_price),
    )
    sample_grid = toolstr.create_grid(sample_mode='quadrants', **render_grid)
    result = toolstr.raster_candlesticks(ohlc.rows(), sample_grid, render_grid)
    raster = result['raster']
    color_grid = result['color_grid']

    as_str = toolstr.render_supergrid(
        raster,
        char_dict='quadrants',
        color_grid=color_grid,
        color_map=toolstr.candlestick_color_map,
    )

    console = rich.console.Console(theme=rich.theme.Theme(inherit=False))

    styles = cli.get_cli_styles()
    tick_label_style = 'bold'
    chrome_style = '#888888'

    y_axis = toolstr.render_y_axis(
        grid=render_grid,
        tick_label_style=tick_label_style,
        chrome_style=chrome_style,
    )
    y_axis_width = rich.text.Text.from_markup(y_axis.split('\n')[0]).cell_len
    graph = toolstr.concatenate_blocks([y_axis, as_str])

    formatter = functools.partial(
        toolstr.format_timestamp,
        representation='TimestampDate',
    )
    x_axis = toolstr.render_x_axis(
        grid=render_grid,
        formatter=formatter,
        tick_label_style=tick_label_style,
        chrome_style=chrome_style,
    )
    x_axis = toolstr.indent_block(x_axis, indent=y_axis_width)

    # compute volume
    if not no_volume:
        max_volume = ohlc['volume'].max()
        if max_volume is None:
            max_volume = 100.0
        ymax = float(max_volume) * 1.1  # type: ignore
        volume_render_grid = toolstr.create_grid(
            n_rows=5,
            n_columns=n_candles * 2,
            xmin=render_grid['xmin'],
            xmax=render_grid['xmax'],
            ymin=0 - ymax / 9,
            ymax=ymax,
        )
        volume_sample_grid = toolstr.create_grid(
            sample_mode='quadrants',
            **volume_render_grid,
        )
        volume_raster = toolstr.raster_bar_chart(
            values=ohlc['volume'].to_list(),
            grid=volume_sample_grid,
            bar_width=1,
            bar_gap=3,
            start_gap=1,
        )
        volume_y_axis = toolstr.render_y_axis(
            grid=volume_render_grid,
            n_ticks=1,
            tick_label_style=tick_label_style,
            chrome_style=chrome_style,
        )

    # wait for metadata
    metadata = await metadata_task

    # print output
    token0 = metadata['x_symbol']
    token1 = metadata['y_symbol']
    toolstr.print_text_box(
        metadata['x_symbol'] + '-' + metadata['y_symbol'] + ' Uniswap V2 Pool',
        style=styles['title'],
    )
    cli.print_bullet(key='pool address', value=pool)
    cli.print_bullet(key='each candle', value=candle_timescale)
    cli.print_bullet(key='n_candles', value=n_candles)
    if invert:
        cli.print_bullet(key='price units', value=str(token1) + ' per ' + str(token0))
    else:
        cli.print_bullet(key='price units', value=str(token0) + ' per ' + str(token1))
    if not no_volume:
        cli.print_bullet(key='volume units', value=token0)
    print()
    console.print(graph)

    if not no_volume:
        volume_bars_str = toolstr.render_supergrid(
            volume_raster, char_dict='quadrants'
        )
        volume_graph = toolstr.concatenate_blocks(
            [volume_y_axis, volume_bars_str]
        )
        toolstr.print(volume_graph)

    # print x axis
    console.print(x_axis)



from __future__ import annotations

import typing

from ctc import spec
from .. import format_utils
from .. import hash_utils
from . import secp256k1_utils

if typing.TYPE_CHECKING:
    from typing_extensions import Literal


def sign_text_message(
    message: str,
    *,
    private_key: spec.Data,
    mode: Literal['eth_sign', 'personal_sign'],
    chain_id: int | None = None,
) -> tuple[int, int, int]:
    """sign text message using private key"""

    message_hash = _hash_text_message(
        message=message,
        mode=mode,
    )

    return sign_message_hash(
        message_hash=message_hash,
        private_key=private_key,
        chain_id=chain_id,
    )


def sign_data_message(
    message: spec.Data,
    *,
    private_key: spec.Data,
    mode: Literal['eth_sign', 'personal_sign'],
    chain_id: int | None = None,
) -> tuple[int, int, int]:
    """sign data message using private key"""

    message_hash = _hash_data_message(
        message=message,
        mode=mode,
    )

    return sign_message_hash(
        message_hash=message_hash,
        private_key=private_key,
        chain_id=chain_id,
    )


def sign_message_hash(
    message_hash: spec.Data,
    *,
    private_key: spec.Data,
    chain_id: int | None = None,
) -> tuple[int, int, int]:
    """sign message hash using private key"""

    message_hash = format_utils.to_binary(message_hash)

    # compute signature
    private_binary = format_utils.to_binary(private_key)
    v, r, s = secp256k1_utils.ecdsa_raw_sign(
        message_hash,
        priv=private_binary,
    )

    # alter v with chain_id
    if chain_id is not None:
        v = v - 27 + chain_id * 2 + 35

    return v, r, s


def _hash_text_message(
    message: str,
    mode: Literal['eth_sign', 'personal_sign'],
) -> bytes:

    # add prefix
    if mode == 'eth_sign':
        full_message = message
    elif mode == 'personal_sign':
        full_message = (
            '\x19Ethereum Signed Message:\n'
            + str(len(message.encode()))
            + message
        )
    else:
        raise Exception('unrecognized signing mode: ' + str(mode))

    # compute message hash
    message_hash = hash_utils.keccak_text(full_message, output_format='binary')

    return message_hash


def _hash_data_message(
    message: spec.Data,
    mode: Literal['eth_sign', 'personal_sign'],
) -> bytes:

    message = format_utils.to_binary(message)

    # add prefix
    if mode == 'eth_sign':
        full_message = message
    elif mode == 'personal_sign':
        prefix = '\x19Ethereum Signed Message:\n' + str(len(message))
        full_message = prefix.encode() + message
    else:
        raise Exception('unrecognized signing mode: ' + str(mode))

    # compute message hash
    message_hash = hash_utils.keccak(full_message, output_format='binary')

    return message_hash



from __future__ import annotations

import typing

from ctc import spec


erc721_collections = {
    # 'crytpokitties': '0x06012c8cf97bead5deae237070f9587f8e7a266d',
    # 'cryptopunks': '0xb47e3cd837ddf8e4c57f05d70ab865de6e193bbb',  # not erc721
    'bayc': '0xbc4ca0eda7647a8ab7c2061c2e118a18a936f13d',
    'mayc': '0x60e4d786628fea6478f785a6d7e704777c86a7c6',
    'milady': '0x5af0d9827e0c53e4799bb226655a1de152a425a5',
    'ens': '0x57f1887a8bf19b14fc0df6fd9b2acc9af147ea85',
    'artbocks': '0xa7d8d9ef8d8ce8992df33d8b8cf4aebabd5bd270',
    'azuki': '0xed5af388653567af2f388e6224dc7c4b3241c544',
    'coolcats': '0x1a92f7381b9f03921564a437210bb9396471050c',
    'uniswapv3position': '0xc36442b4a4522e871399cd717abdd847ab11fe88',
    'loot': '0xff9c1b15b16263c61d017ee9f65c50e4ae0113d7',
    'nouns': '0x9c8ff314c9bc7f6e59a9d9225fb22946427edc03',
    'pudgypenguins': '0xbd3531da5cf5857e7cfaa92426877b022e612cf8',
}


erc721_function_abis: typing.Mapping[str, spec.FunctionABI] = {
    'balanceOf': {
        'inputs': [
            {
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
        ],
        'name': 'balanceOf',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'totalSupply': {
        'inputs': [],
        'name': 'totalSupply',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'ownerOf': {
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'tokenId',
                'type': 'uint256',
            },
        ],
        'name': 'ownerOf',
        'outputs': [
            {
                'internalType': 'address',
                'name': '',
                'type': 'address',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'getApproved': {
        'inputs': [
            {
                'internalType': 'uint256',
                'name': 'tokenId',
                'type': 'uint256',
            },
        ],
        'name': 'getApproved',
        'outputs': [
            {
                'internalType': 'address',
                'name': '',
                'type': 'address',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'isApprovedForAll': {
        'inputs': [
            {
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
            {
                'internalType': 'address',
                'name': 'operator',
                'type': 'address',
            },
        ],
        'name': 'isApprovedForAll',
        'outputs': [
            {
                'internalType': 'bool',
                'name': '',
                'type': 'bool',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
}

erc721_event_abis: typing.Mapping[str, spec.EventABI] = {
    'Approval': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'approved',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'uint256',
                'name': 'tokenId',
                'type': 'uint256',
            },
        ],
        'name': 'Approval',
        'type': 'event',
    },
    'Transfer': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'from',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'to',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'uint256',
                'name': 'tokenId',
                'type': 'uint256',
            },
        ],
        'name': 'Transfer',
        'type': 'event',
    },
    'ApprovalForAll': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'owner',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'operator',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'bool',
                'name': 'approved',
                'type': 'bool',
            },
        ],
        'name': 'ApprovalForAll',
        'type': 'event',
    },
}



from __future__ import annotations

import toolcli

from ctc.protocols import ens_utils
from ctc import evm
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_exists_command,
        'help': 'output whether ENS name exists',
        'args': [
            {'name': 'name', 'help': 'ENS name'},
            {'name': '--block', 'help': 'block number'},
        ],
        'examples': [
            'vitalik.eth',
            'vitalik.eth --block 14000000',
            'fake-vitalik.eth',
        ],
    }


async def async_exists_command(
    name: str, block: spec.BlockNumberReference
) -> None:

    if block is not None:
        block = evm.standardize_block_number(block)
    exists = await ens_utils.async_record_exists(name, block=block)

    print(exists)



"""WIP
Desired Information
- token creation date
- holders
    - top token holders
    - number of holders
- token dex summary on major exchanges
    - volume
    - liquidity and liquidity depth
- token CEX activity and inflow/outflow
- how each of these values change over time
- available oracles
"""

from __future__ import annotations

import toolcli

from ctc import evm


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_erc20_command,
        'help': 'display information about ERC20',
        'args': [
            {
                'name': 'erc20',
                'help': 'ERC20 address or symbol',
            },
        ],
        'examples': [
            'DAI',
            '0x6b175474e89094c44da98b954eedeac495271d0f',
        ],
    }


async def async_erc20_command(erc20: str) -> None:
    await evm.async_print_erc20_summary(erc20)



from __future__ import annotations

import typing

import toolsql

from ctc import spec
from ctc.db import schema_utils
from . import table_block_transaction_queries


async def async_select_block_transactions(
    block_number: int,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> typing.Sequence[spec.DBTransaction] | None:
    result = await async_select_blocks_transactions(
        block_numbers=[block_number],
        conn=conn,
        context=context,
    )
    if result is None:
        return None
    transactions, manifest = result
    if len(manifest) > 0:
        return transactions
    else:
        return None


async def async_select_blocks_transactions(
    *,
    block_numbers: typing.Sequence[int] | None = None,
    start_block: int | None = None,
    end_block: int | None = None,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> tuple[typing.Sequence[spec.DBTransaction], typing.Sequence[int]] | None:

    blocks_present = await table_block_transaction_queries.async_select_block_transaction_queries(
        block_numbers=block_numbers,
        start_block=start_block,
        end_block=end_block,
        conn=conn,
        context=context,
    )
    if blocks_present is None or len(blocks_present) == 0:
        return None

    table = schema_utils.get_table_schema('transactions', context=context)
    columns = [
        column['name']
        for column in table['columns']
        if column['name'] != 'access_list'
    ]

    if block_numbers is not None:
        transactions: typing.Sequence[
            spec.DBTransaction
        ] = await toolsql.async_select(  # type: ignore
            conn=conn,
            table=table,
            where_in={'block_number': blocks_present},
            columns=columns,
        )

    else:
        if start_block is not None:
            where_gte = {'block_number': start_block}
        else:
            where_gte = None
        if end_block is not None:
            where_lte = {'block_number': end_block}
        else:
            where_lte = None
        transactions = await toolsql.async_select(  # type: ignore
            conn=conn,
            table=table,
            where_lte=where_lte,
            where_gte=where_gte,
            columns=columns,
        )

    return transactions, blocks_present




from __future__ import annotations

import typing
from typing_extensions import TypedDict


class RawLog(TypedDict):
    removed: bool
    logIndex: int
    transactionIndex: int
    transactionHash: str
    blockHash: str
    blockNumber: int
    address: str
    data: str
    topics: typing.List[str]


class PendingRawLog(TypedDict):
    # many log fields are nullable if a log is pending
    removed: bool
    logIndex: typing.Union[None, int]
    transactionIndex: typing.Union[None, int]
    transactionHash: typing.Union[None, str]
    blockHash: typing.Union[None, str]
    blockNumber: typing.Union[None, int]
    address: str
    data: str
    topics: typing.List[str]


NormalizedLog = typing.Dict[str, typing.Any]




from __future__ import annotations

import time
import typing

import toolcli
import toolstr
import tooltime

from ctc import config
from ctc import evm
from ctc import rpc
from ctc import spec
from ctc.cli import cli_utils
from ctc.protocols import chainlink_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_chainlink_command,
        'help': 'output Chainlink feed data',
        'args': [
            {'name': 'feed', 'nargs': '+', 'help': 'name or address of feed'},
            {
                'name': ['--verbose', '-v'],
                'help': 'show additional information',
                'action': 'store_true',
            },
            {
                'name': '--blocks',
                'help': 'block range of datapoints',
            },
            {
                'name': '--time',
                'dest': 'timelength',
                'help': 'historical duration to show feed data',
            },
            {
                'name': '--export',
                'default': 'stdout',
                'help': 'file path for output (.json or .csv)',
            },
            {
                'name': '--overwrite',
                'action': 'store_true',
                'help': 'specify that output path can be overwritten',
            },
            {'name': '--provider', 'help': 'rpc provider name or url'},
            {'name': '--network', 'help': 'network name or chain_id'},
            {
                'name': '--all-fields',
                'action': 'store_true',
                'help': 'include all output fields',
            },
            {
                'name': '--interpolate',
                'help': 'interpolate all blocks in range',
                'action': 'store_true',
            },
        ],
        'examples': [
            'DAI_USD',
            '0xaed0c38402a5d19df6e4c03f4e2dced6e29c1ee9',
            'DAI_USD --blocks 14000000:14001000',
        ],
    }


async def async_chainlink_command(
    *,
    feed: typing.Sequence[str],
    verbose: bool,
    blocks: str | None,
    timelength: str,
    export: str,
    overwrite: bool,
    provider: str | None,
    network: spec.NetworkReference | None,
    all_fields: bool | None,
    interpolate: bool,
) -> None:

    context = config.create_user_input_context(
        provider=provider,
        network=network,
    )

    feed_str = '_'.join(feed)

    feed_str = await evm.async_resolve_address(feed_str, context=context)

    if timelength is not None:
        timelength_seconds = tooltime.timelength_to_seconds(timelength)
        start_block: int | None = await evm.async_get_block_of_timestamp(
            time.time() - timelength_seconds,
            mode='<=',
            context=context,
        )
    else:
        start_block = None

    if all_fields:
        fields: typing.Literal['full', 'answer'] = 'full'
    else:
        fields = 'answer'

    if blocks is None:
        await chainlink_utils.async_print_feed_summary(
            feed=feed_str,
            verbose=verbose,
            start_block=start_block,
            context=context,
        )
    else:
        feed_address = await chainlink_utils.async_resolve_feed_address(
            feed=feed_str,
            context=context,
        )
        name = await rpc.async_eth_call(
            feed_address,
            function_abi=chainlink_utils.feed_function_abis['description'],
            context=context,
        )
        toolstr.print_text_box('Chainlink Feed: ' + name)
        print('- feed address')
        print('- feed:', feed_str)
        print('- fields:', fields)
        print('- output:', export)

        if blocks is not None:
            start_block, end_block = await cli_utils.async_parse_block_range(
                blocks,
                context=context,
            )
        else:
            start_block = None
            end_block = None
        # block_kwargs = {'blocks': resolved_blocks}
        # print('- n_blocks:', len(resolved_blocks))
        if start_block is not None:
            print('- start_block:', start_block)
        if end_block is not None:
            print('- end_block:', end_block)

        feed_data = await chainlink_utils.async_get_feed_data(
            feed_str,
            fields=fields,
            start_block=start_block,
            end_block=end_block,
            interpolate=interpolate,
            context=context,
        )
        df = feed_data

        print()
        cli_utils.output_data(df, output=export, overwrite=overwrite, raw=True)



from __future__ import annotations

import toolcli

from ctc.protocols import aave_v2_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_aave_addresses_command,
        'help': 'output aave-related addresses',
        'args': [
            {
                'name': ['--verbose', '-v'],
                'help': 'show additional addresses',
                'action': 'store_true',
            },
            {
                'name': '--block',
                'help': 'block to get addresses of',
            },
        ],
        'examples': [
            '',
            '--verbose',
            '--block 15000000',
        ],
    }


async def async_aave_addresses_command(
    verbose: bool,
    block: str | None,
) -> None:
    await aave_v2_utils.async_print_aave_addresses(
        verbose=verbose,
        block=block,
    )



from __future__ import annotations

import os
import typing

from ctc import spec


def get_config_validators() -> (
    typing.Mapping[str, None | typing.Callable[..., None]]
):
    return {
        'config_spec_version': validate_config_spec_version,
        'data_dir': validate_data_dir,
        'networks': validate_networks,
        'providers': validate_providers,
        'default_network': validate_default_network,
        'default_providers': validate_default_providers,
        'db_configs': validate_db_configs,
        'log_rpc_calls': None,
        'log_sql_queries': None,
        'cli_color_theme': None,
        'cli_chart_charset': None,
        'context_cache_rules': validate_context_cache_rules,
    }


def get_config_base_types() -> (
    typing.Mapping[
        str, typing.Type[typing.Any] | tuple[typing.Type[typing.Any], ...]
    ]
):
    return {
        'config_spec_version': str,
        'data_dir': str,
        'networks': dict,
        'providers': dict,
        'default_network': int,
        'default_providers': dict,
        'db_configs': dict,
        'log_rpc_calls': bool,
        'log_sql_queries': bool,
        'cli_color_theme': dict,
        'cli_chart_charset': str,
        'context_cache_rules': list,
    }


def validate_config(config: typing.Mapping[typing.Any, typing.Any]) -> None:
    """raise spec.ConfigInvalid if config is not valid"""

    config_validators = get_config_validators()
    config_base_types = get_config_base_types()

    # check that all required keys are present
    for key in spec.config_keys:
        if key not in config:
            raise spec.ConfigInvalid('config does not specify key: ' + str(key))

    # check that each entry is valid
    for key, value in config.items():
        # check that key is allowed
        if key not in spec.config_keys:
            raise spec.ConfigInvalid('key not allowed in config: ' + str(key))

        # check value type
        if key not in config_base_types:
            raise Exception('type unknown for config field: ' + str(key))
        if not isinstance(value, config_base_types[key]):
            message = 'invalid type for ' + str(key) + ': ' + str(value)
            raise spec.ConfigInvalid(message)

    # check additional key validations
    for key, value in config.items():
        key_validator = config_validators.get(key)
        if key_validator is not None:
            key_validator(value=value, config=config)

    validate_non_overlapping_identifiers(config)


def is_valid_config(config: typing.Mapping[typing.Any, typing.Any]) -> bool:
    try:
        validate_config(config)
        return True
    except spec.ConfigInvalid:
        return False


#
# # individual validators
#


def validate_config_spec_version(
    value: typing.Any, config: typing.Mapping[typing.Any, typing.Any]
) -> None:
    pieces = value.split('.')
    if len(pieces) != 3:
        raise spec.ConfigInvalid('invalid config_spec_version: ' + str(value))
    if not all(piece.isnumeric() for piece in pieces[:2]):
        raise spec.ConfigInvalid('invalid config_spec_version: ' + str(value))

    if pieces[0] != '0' or pieces[1] != '3':
        raise spec.ConfigInvalid('config must use a version 0.3.x')


def validate_data_dir(
    value: typing.Any, config: typing.Mapping[typing.Any, typing.Any]
) -> None:
    abspath = os.path.abspath(os.path.expanduser(value))
    if value != abspath:
        raise spec.ConfigInvalid(
            'data_dir path should be absolute: ' + str(value)
        )


def validate_networks(
    value: typing.Any, config: typing.Mapping[typing.Any, typing.Any]
) -> None:
    config_keys = {'name', 'chain_id', 'block_explorer'}
    for chain_id, network_metadata in value.items():
        if not isinstance(chain_id, int):
            raise spec.ConfigInvalid(
                'networks should be identified by int chain_id'
            )
        if set(network_metadata.keys()) != config_keys:
            raise spec.ConfigInvalid(
                'network_metadata should specify name, chain_id, and block_explorer, instead got: '
                + str(list(network_metadata.keys()))
            )

        name = network_metadata['name']
        block_explorer = network_metadata['block_explorer']
        if name is not None and not isinstance(name, str):
            raise spec.ConfigInvalid('network name is not a str')
        if network_metadata['chain_id'] != chain_id:
            raise spec.ConfigInvalid('chain_id does not match')
        if block_explorer is not None and not isinstance(block_explorer, str):
            raise spec.ConfigInvalid('block_explorer is not a str')


def validate_providers(
    value: typing.Any, config: typing.Mapping[typing.Any, typing.Any]
) -> None:
    for provider_name, provider in value.items():
        if not isinstance(provider_name, str):
            raise spec.ConfigInvalid('provider name should be a str')
        if 'name' not in provider or provider_name != provider['name']:
            raise spec.ConfigInvalid('provider name does not match')
        validate_provider(provider, config)


def validate_provider(
    provider: typing.Any, config: typing.Mapping[typing.Any, typing.Any]
) -> None:
    if not isinstance(provider, dict):
        raise Exception('provider is not a dict')

    provider_keys = set(spec.provider_keys)
    if set(provider.keys()) != provider_keys:
        raise spec.ConfigInvalid(
            'provider should have keys: '
            + str(provider_keys)
            + ', it has keys: '
            + str(provider.keys())
        )

    url = provider['url']
    name = provider['name']
    network = provider['network']
    protocol = provider['protocol']
    session_kwargs = provider['session_kwargs']
    chunk_size = provider['chunk_size']

    if not isinstance(url, str):
        raise spec.ConfigInvalid('provider url must be a str')
    if not isinstance(name, str):
        raise spec.ConfigInvalid('provider name must be a str')
    if not isinstance(network, int):
        raise spec.ConfigInvalid('provider networks should be an int chain_id')
    if network not in config['networks']:
        raise spec.ConfigInvalid('provider network not in network entries')
    if protocol != 'http':
        raise spec.ConfigInvalid('only http supported')
    if not isinstance(session_kwargs, dict):
        raise spec.ConfigInvalid('session_kwargs must be a dict')
    if chunk_size is not None and not isinstance(chunk_size, int):
        raise spec.ConfigInvalid('chunk_size is not int')

    if not url.startswith('http'):
        raise spec.ConfigInvalid(
            'http provider url must start with "http://" or "https://"'
        )


def validate_default_network(
    value: typing.Any, config: typing.Mapping[typing.Any, typing.Any]
) -> None:
    if value is None:
        return

    # check that value is present in networks
    networks = config['networks']
    if value not in networks:
        raise spec.ConfigInvalid('default_network not in networks entries')


def validate_default_providers(
    value: typing.Any, config: typing.Mapping[typing.Any, typing.Any]
) -> None:
    networks = config['networks']
    providers = config['providers']
    for chain_id, provider_name in value.items():
        if chain_id not in networks:
            raise spec.ConfigInvalid(
                'chain_id in default_providers not present in network entries'
            )
        if provider_name not in providers:
            raise spec.ConfigInvalid(
                'provider in default_providers not present in provider entries'
            )


def validate_db_configs(
    value: typing.Any, config: typing.Mapping[typing.Any, typing.Any]
) -> None:
    if not isinstance(value, dict):
        raise spec.ConfigInvalid('invalid format for db_config')
    if len(value) == 0:
        raise spec.ConfigInvalid('no db specified')
    for db_name, db_config in value.items():
        if 'dbms' not in db_config:
            raise Exception('dbms not specified for db_name')
        if db_config['dbms'] == 'sqlite':
            if set(db_config.keys()) != {'dbms', 'path'}:
                raise spec.ConfigInvalid(
                    'db config should have keys dbms and path'
                )
        elif db_config['dbms'] == 'postgresql':
            if set(db_config.keys()) != {
                'dbms',
                'database',
                'hostname',
                'port',
                'username',
                'password',
            }:
                raise spec.ConfigInvalid(
                    'db config should have keys dbms and path'
                )
        else:
            raise spec.ConfigInvalid(
                'unrecognized dbms: ' + str(db_config['dbms'])
            )


def validate_context_cache_rules(
    value: typing.Any, config: typing.Mapping[typing.Any, typing.Any]
) -> None:
    if not isinstance(value, list):
        raise Exception('context_cache_rules must be a list of rules')

    for rule in value:
        if not isinstance(rule, dict):
            raise Exception('context cache rule must be a dict')
        keys = set(rule.keys())
        extra_keys = keys.difference({'backend', 'read', 'write', 'filter'})
        if len(extra_keys) > 0:
            raise Exception('extra keys in rule: ' + str(extra_keys))

        if 'backend' in rule:
            backend = rule.get('backend')
            assert backend in config['db_configs']
        if 'read' in rule:
            read = rule.get('read')
            assert isinstance(read, bool)
        if 'write' in rule:
            write = rule.get('write')
            assert isinstance(write, bool)
        if 'filter' in rule:
            filter = rule.get('filter')
            assert isinstance(filter, dict)
            if 'chain_id' in filter:
                assert filter['chain_id'] in config['networks']
            if 'schema' in filter:
                assert filter['schema'] in spec.schema_names
            if 'backend' in filter:
                assert filter['backend'] in config['db_configs']


def validate_non_overlapping_identifiers(
    config: typing.Mapping[typing.Any, typing.Any]
) -> None:
    providers = config.get('providers')
    if not isinstance(providers, dict):
        raise Exception('invalid format for providers')
    provider_names = {
        provider['name'] for provider_name, provider in providers.items()
    }
    assert len(provider_names) == len(providers), 'non-unique provider names'

    networks = config.get('networks', {})
    if not isinstance(networks, dict):
        raise Exception('invalid format for networks')
    network_names = {metadata['name'] for metadata in networks.values()}
    assert len(network_names) == len(networks), 'non-unique network names'

    db_configs = config.get('db_configs', {})
    if not isinstance(db_configs, dict):
        raise Exception('invalid format for db_configs')
    db_names = {db_name for db_name in db_configs.keys()}
    assert len(db_names) == len(db_configs), 'non-unique database names'

    if len(provider_names.intersection(network_names)) > 0:
        raise Exception('naming conflict between providers and networks')
    if len(provider_names.intersection(db_names)) > 0:
        raise Exception('naming conflict between providers and databases')
    if len(db_names.intersection(network_names)) > 0:
        raise Exception('naming conflict between databases and networks')




import toolsql

from ctc import db

import conftest


example_data = [
    {
        'address': '0x956f47f50a910163d8bf957cf5846d573e7f87ca',
        'block_number': 12125705,
    },
    {
        'address': '0xc7283b66eb1eb5fb86327f08e1b5816b0720212b',
        'block_number': 12125705,
    },
    {
        'address': '0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48',
        'block_number': 6082465,
    },
]


async def test_create_creation_blocks_crud():
    db_config = conftest.get_test_db_config()
    db_schema = db.get_prepared_schema(
        schema_name='contract_creation_blocks',
        context=dict(network='ethereum'),
    )
    toolsql.create_db(
        db_config=db_config,
        db_schema=db_schema,
        if_not_exists=True,
        confirm=True,
    )

    # insert data
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            await db.async_upsert_contract_creation_block(
                conn=conn, **datum
            )

    # get data individually
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            stored_block = await db.async_select_contract_creation_block(
                conn=conn,
                address=datum['address'],
            )
            assert stored_block == datum['block_number']

    # delete entries one by one
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            await db.async_delete_contract_creation_block(
                conn=conn,
                address=datum['address'],
            )

    # ensure all entries deleted
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            block = await db.async_select_contract_creation_block(
                conn=conn,
                address=datum['address'],
            )
            assert block is None




from __future__ import annotations

import typing

from ctc import spec
from .. import format_utils

if typing.TYPE_CHECKING:
    from typing_extensions import Literal


def pack_signature_vrs(
    *vrs: spec.Data,
    mode: Literal['transaction', 'ecdsa'],
    v: spec.Data | None = None,
    r: spec.Data | None = None,
    s: spec.Data | None = None,
) -> str:
    """convert signature to singleton hdex data representation"""

    if len(vrs) == 3:
        v, r, s = vrs

    if v is not None and r is not None and s is not None:
        v_bytes = format_utils.to_binary(v, n_bytes=1)
        r_bytes = format_utils.to_binary(r, n_bytes=32)
        s_bytes = format_utils.to_binary(s, n_bytes=32)
    else:
        raise Exception('must specify v, r, and s')

    if mode == 'ecdsa':
        signature = r_bytes + s_bytes + v_bytes
    elif mode == 'transaction':
        delimiter = bytes.fromhex('a0')
        signature = v_bytes + delimiter + r_bytes + delimiter + s_bytes
    else:
        raise Exception('unknown vrs packing mode: ' + str(mode))

    return format_utils.to_hex(signature)


def unpack_signature_vrs(signature: spec.Signature) -> tuple[int, int, int]:
    """convert signature to triplet VRS representation"""

    if isinstance(signature, tuple):
        v, r, s = signature

    else:
        bytes_signature = format_utils.to_binary(signature)

        if len(bytes_signature) == 65:
            r = bytes_signature[:32]
            s = bytes_signature[32:64]
            v = bytes_signature[64:]
        elif len(bytes_signature) == 67:
            v = bytes_signature[:1]
            r = bytes_signature[2:34]
            s = bytes_signature[35:]
        else:
            raise Exception('signature format unrecognized')

    return (
        format_utils.binary_convert(v, 'integer'),
        format_utils.binary_convert(r, 'integer'),
        format_utils.binary_convert(s, 'integer'),
    )



from __future__ import annotations

import typing

from ctc import evm
from ctc import spec
from .. import function_abi_utils
from .. import event_abi_utils


def get_contract_abi_by_selectors(
    contract_abi: spec.ContractABI,
) -> typing.Mapping[str, spec.ContractABIEntry]:
    """return a mapping from function/event selectors to function/event abis

    omits constructor, receive, and fallback functions
    """
    by_selectors: typing.MutableMapping[str, spec.ContractABIEntry] = {}
    for item in contract_abi:
        if item['type'] == 'function':
            function_selector = function_abi_utils.get_function_selector(item)
            by_selectors[function_selector] = item
        elif item['type'] == 'event':
            event_hash = event_abi_utils.get_event_hash(item)
            by_selectors[event_hash] = item
        elif item['type'] in ['error', 'constructor', 'receive', 'fallback']:
            pass
        else:
            raise Exception('unknown item type in contract abi')
    return by_selectors


def print_contract_abi(
    contract_abi: spec.ContractABI,
    *,
    max_width: int | None = None,
    verbose: bool | int = False,
    read_write: bool = False,
) -> None:
    """print summary of contract ABI"""

    print_contract_abi_functions(
        contract_abi=contract_abi,
        max_width=max_width,
        verbose=verbose,
        read_write=read_write,
    )
    print()
    print()
    print_contract_abi_events(
        contract_abi=contract_abi,
        max_width=max_width,
        verbose=verbose,
    )


def print_contract_abi_functions(
    contract_abi: spec.ContractABI,
    *,
    max_width: int | None = None,
    verbose: bool | int = False,
    title: str = 'Contract ABI Functions',
    read_write: bool = False,
) -> None:
    """print summary of contract ABI functions"""

    import toolstr
    from ctc import cli

    styles = cli.get_cli_styles()

    if read_write:
        read_color = styles['description']
        write_color = 'white'

    functions = evm.get_function_abis(contract_abi)

    if read_write:
        functions = sorted(
            functions,
            key=function_abi_utils.is_function_read_only,
            reverse=True,
        )

    toolstr.print_text_box(title, style=styles['title'])
    print()
    rows = []
    for i, function in enumerate(functions):

        if len(function['outputs']) == 0:
            output_str = '-'
        else:
            output_str_list = [
                (item['type'] + ' ' + item['name']).strip()
                for item in function['outputs']
            ]
            output_str = ', '.join(output_str_list)
            output_str = output_str.strip()
        output_str = output_str.strip()
        if len(output_str) == 0:
            output_str = '-'

        inputs = function.get('inputs', [])
        if len(inputs) == 0:
            inputs = [{'name': '-', 'type': '-'}]

        name = function['name']
        if read_write:
            if function_abi_utils.is_function_read_only(function):
                # read
                name = toolstr.add_style(name, read_color)
            else:
                # write
                name = toolstr.add_style(name, write_color)

        row = [
            function_abi_utils.get_function_selector(function),
            name,
        ]

        if verbose:
            input_items = []
            for item in function.get('inputs', []):
                item_str = item.get('type', '') + ' ' + item.get('name', '')
                item_str = item_str.strip()
                input_items.append(item_str)
            input_str = '\n'.join(input_items)
            if input_str == '':
                input_str = '-'

            row.append(input_str)
            row.append(output_str)
        else:
            row.append(', '.join(item['type'] for item in inputs))
            row.append(output_str)
        rows.append(row)

    name_column = 'name'
    if read_write:
        name_column += (
            ' ('
            + toolstr.add_style('r', read_color)
            + '/'
            + toolstr.add_style('w', write_color)
            + ')'
        )

    labels = [
        'selector',
        name_column,
        'inputs',
        'outputs',
    ]
    if verbose:
        max_column_widths: typing.Mapping[int | str, int] | None = None
    else:
        max_column_widths = {'inputs': 25, 'outputs': 25}

    if verbose:
        toolstr.print_multiline_table(
            rows,
            add_row_index=True,
            labels=labels,
            max_column_widths=max_column_widths,
            compact=4,
            max_table_width=max_width,
            border=styles['comment'],
            label_style=styles['title'],
            column_styles={
                'selector': styles['metavar'],
                'inputs': styles['option'],
                'outputs': styles['option'],
            },
        )
    else:
        toolstr.print_table(
            rows,
            add_row_index=True,
            labels=labels,
            max_column_widths=max_column_widths,
            max_table_width=max_width,
            border=styles['comment'],
            label_style=styles['title'],
            column_styles={
                'selector': styles['metavar'],
                'inputs': styles['option'],
                'outputs': styles['option'],
            },
        )

    print()
    has_constructor = any(
        item['type'] == 'constructor' for item in contract_abi
    )
    has_receive = any(item['type'] == 'receive' for item in contract_abi)
    has_fallback = any(item['type'] == 'fallback' for item in contract_abi)

    print()
    toolstr.print_text_box('Special Functions Present', style=styles['title'])
    print()
    special_rows = [
        ('constructor', has_constructor),
        ('receive', has_receive),
        ('fallback', has_fallback),
    ]
    toolstr.print_table(
        special_rows,
        labels=['function', 'present'],
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'function': styles['option'],
            'present': styles['description'],
        },
        # indent=4,
    )


def print_contract_abi_events(
    contract_abi: spec.ContractABI,
    *,
    max_width: int | None = None,
    verbose: bool | int = False,
    title: str = 'Contract ABI Events',
) -> None:
    """print summary of contract ABI events"""

    import toolstr
    from ctc import cli

    styles = cli.get_cli_styles()

    events = evm.get_event_abis(contract_abi)
    toolstr.print_text_box(title, style=styles['title'])
    print()
    rows = []

    for i, event_abi in enumerate(events):

        name = event_abi['name']
        row = [name]
        input_cell = []
        indexed_cell = []
        for item in event_abi.get('inputs', []):
            subitems = [item.get('type'), item.get('name')]
            subitems_str = [
                subitem for subitem in subitems if subitem is not None
            ]
            input_str = ' '.join(subitems_str)
            input_cell.append(input_str)
            if item.get('indexed'):
                indexed_cell.append('✓')
            else:
                indexed_cell.append('')
        row.append('\n'.join(input_cell))
        row.append('\n'.join(indexed_cell))

        event_hash = event_abi_utils.get_event_hash(event_abi)
        if verbose:
            row.append(event_hash)
        else:
            row.append(event_hash[:6] + '...' + event_hash[-4:])

        rows.append(row)
    labels = ['name', 'inputs', 'indexed', 'event_hash']

    toolstr.print_multiline_table(
        rows,
        add_row_index=True,
        missing_columns='fill',
        labels=labels,
        vertical_justify='top',
        compact=4,
        column_justify={'indexed': 'center'},
        max_table_width=max_width,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'event_hash': styles['metavar'],
            'inputs': styles['option'],
            'name': styles['description'],
            'indexed': styles['description'],
        },
    )


def contract_abi_to_dataframe(
    contract_abi: spec.ContractABI,
    human_readable: bool,
) -> spec.DataFrame:
    """create DataFrame representation of contract ABI"""

    import copy
    import polars as pl

    contract_abi = copy.deepcopy(contract_abi)
    for entry in typing.cast(
        typing.List[typing.Dict[str, typing.Any]], contract_abi
    ):

        if human_readable:

            if 'name' not in entry:
                entry['name'] = ''

            if 'inputs' in entry:
                inputs = []
                for input in entry['inputs']:
                    input_str = ''
                    input_str += input.get('type', '')
                    input_name = input.get('name', '')
                    if input_name is not None and input_name != '':
                        input_str += ' ' + input_name
                    inputs.append(input_str)
                entry['inputs'] = ', '.join(inputs)
            else:
                entry['inputs'] = ''

            if 'outputs' in entry:
                outputs = []
                for output in entry['outputs']:
                    output_str = ''
                    output_str += output.get('type')
                    output_name = output.get('name')
                    if output_name is not None and output_name != '':
                        output_str += ' ' + output_name
                    outputs.append(output_str)
                entry['outputs'] = ', '.join(outputs)
            else:
                entry['outputs'] = ''

            if 'anonymous' not in entry:
                entry['anonymous'] = ''

            if 'stateMutability' not in entry:
                entry['stateMutability'] = ''

    df = pl.DataFrame(contract_abi)

    return df




from __future__ import annotations

import typing

from ctc import spec
from . import erc721_events
from . import erc721_spec
from . import erc721_state

if typing.TYPE_CHECKING:
    from typing_extensions import Literal


async def async_get_erc721_owners(
    token: spec.Address,
    *,
    method: Literal['transfers', 'calls'] | None = None,
    token_ids: typing.Sequence[int] | None = None,
    context: spec.Context = None,
) -> typing.Mapping[int, spec.Address]:
    """return owners of erc721 token

    can multiple methods:
    - transfers: compute owners based on erc721 transfers
    - calls: use eth_call's on enumerated list of all token_id's
    """

    if method is None:
        if token_ids is not None:
            method = 'calls'
        else:
            method = 'transfers'

    if method == 'transfers':
        transfers = await erc721_events.async_get_erc721_transfers(
            token, context=context
        )
        return _get_erc721_owners_from_transfers(transfers, token_ids=token_ids)

    elif method == 'calls':
        return await _async_get_erc721_owners_from_calls(
            token,
            token_ids=token_ids,
            context=context,
        )

    else:
        raise Exception('unknown ownership method')


def _get_erc721_owners_from_transfers(
    transfers: spec.DataFrame,
    *,
    token_ids: typing.Sequence[int] | None = None,
) -> typing.Mapping[int, spec.Address]:

    owner_transfers = (
        transfers[['arg__tokenId', 'arg__to']].groupby('arg__tokenId').last()
    )
    owner_dict = typing.cast(
        typing.Mapping[int, str],
        dict(owner_transfers.rows()),  # type: ignore
    )

    if token_ids is not None:
        filtered_owner_dict = {}
        for token_id in token_ids:
            if token_id not in owner_dict:
                raise Exception('could not find token_id in transfers')
            filtered_owner_dict[token_id] = owner_dict[token_id]
        owner_dict = filtered_owner_dict

    return owner_dict


async def _async_get_erc721_owners_from_calls(
    token: spec.Address,
    token_ids: typing.Sequence[int] | None = None,
    *,
    context: spec.Context = None,
) -> typing.Mapping[int, spec.Address]:
    from ctc import rpc

    if token_ids is None:
        try:
            total_supply = await erc721_state.async_get_erc721_total_supply(
                token,
                context=context,
            )
            token_ids = list(range(total_supply))
        except spec.RpcException:
            raise Exception(
                'this erc721 does not implement totalSupply(), to get owners from calls, need to provide token_ids directly'
            )

    results = await rpc.async_batch_eth_call(
        to_address=token,
        function_abi=erc721_spec.erc721_function_abis['ownerOf'],
        function_parameter_list=[[token_id] for token_id in token_ids],
        context=context,
    )

    return dict(zip(token_ids, results))




from __future__ import annotations

import toolcli
import toolstr

from ctc import evm
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_balance_command,
        'help': 'output an ERC20 balance',
        'args': [
            {'name': 'erc20_address', 'help': 'address of ERC20 token'},
            {'name': 'wallet_address', 'help': 'address of wallet'},
            {'name': '--block', 'help': 'block number'},
            {
                'name': '--raw',
                'action': 'store_true',
                'help': 'whether to normalize balance by ERC20 decimals',
            },
        ],
        'examples': [
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca 0x9928e4046d7c6513326ccea028cd3e7a91c7590a',
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca 0x9928e4046d7c6513326ccea028cd3e7a91c7590a --raw',
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca 0x9928e4046d7c6513326ccea028cd3e7a91c7590a --block 14000000',
        ],
    }


async def async_balance_command(
    *,
    erc20_address: spec.Address,
    wallet_address: spec.Address,
    block: str,
    raw: bool,
) -> None:
    erc20_address = await evm.async_resolve_address(erc20_address, block=block)
    wallet_address = await evm.async_resolve_address(
        wallet_address,
        block=block,
    )
    balance = await evm.async_get_erc20_balance(
        wallet=wallet_address,
        token=erc20_address,
        block=block,
        normalize=(not raw),
    )
    if raw:
        print(balance)
    else:
        print(toolstr.format(balance))



from __future__ import annotations

import typing

import toolcli
import toolstr

from ctc import evm
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_balance_command,
        'help': 'output ETH balance of address',
        'args': [
            {'name': 'address', 'help': 'address of wallet'},
            {'name': '--block', 'help': 'block number'},
            {
                'name': '--raw',
                'action': 'store_true',
                'help': 'skip normalizing balance by 1e18 decimals',
            },
        ],
        'examples': [
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045',
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045 --raw',
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045 --block 14000000',
        ],
    }


async def async_balance_command(
    *,
    address: str,
    block: typing.Optional[spec.BlockNumberReference],
    raw: bool,
) -> None:

    address = await evm.async_resolve_address(address, block=block)

    if block is not None:
        block = await evm.async_block_number_to_int(block)

    balance = await evm.async_get_eth_balance(
        address=address,
        block=block,
        normalize=(not raw),
    )
    if raw:
        print(balance)
    else:
        print(toolstr.format(balance))



from __future__ import annotations

from ...management import active_utils
from ... import query_utils
from . import multischema_block_timestamps_statements
from . import multischema_block_timestamps_search


async_query_block_timestamp = query_utils.wrap_selector_with_connection(
    multischema_block_timestamps_statements.async_select_block_timestamp,
    active_utils.get_active_timestamp_schema,
)

async_query_block_timestamps = query_utils.wrap_selector_with_connection(
    multischema_block_timestamps_statements.async_select_block_timestamps,
    active_utils.get_active_timestamp_schema,
)

async_query_max_block_number = query_utils.wrap_selector_with_connection(
    multischema_block_timestamps_statements.async_select_max_block_number,
    active_utils.get_active_timestamp_schema,
)

async_query_max_block_timestamp = query_utils.wrap_selector_with_connection(
    multischema_block_timestamps_statements.async_select_max_block_timestamp,
    active_utils.get_active_timestamp_schema,
)

async_query_timestamp_block = query_utils.wrap_selector_with_connection(
    multischema_block_timestamps_search.async_select_timestamp_block,
    active_utils.get_active_timestamp_schema,
)

async_query_timestamps_blocks = query_utils.wrap_selector_with_connection(
    multischema_block_timestamps_search.async_select_timestamps_blocks,
    active_utils.get_active_timestamp_schema,
)

async_query_timestamp_block_range = query_utils.wrap_selector_with_connection(
    multischema_block_timestamps_search.async_select_timestamp_block_range,
    active_utils.get_active_timestamp_schema,
)



from .block_gas import *
from .block_timestamps import *
from .blocks import *
from .contract_abis import *
from .contract_creation_blocks import *
from .dex_pools import *
from .erc20_metadata import *
from .events import *
from .schema_versions import *
from .transactions import *




from __future__ import annotations

import toolcli

from ctc.protocols import etherscan_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_etherscan_command,
        'args': [
            {
                'name': 'query',
                'help': 'address / tx / block / ERC20 symbol / ens',
            },
            {
                'name': '--holdings',
                'help': 'open page viewing holdings of address',
                'action': 'store_true',
            },
            {
                'name': '--erc20-transfers',
                'help': 'open page viewing ERC20 transfers of address',
                'action': 'store_true',
            },
            {
                'name': '--internal',
                'help': 'open page viewing internal txs of address',
                'action': 'store_true',
            },
            {
                'name': '--holders',
                'help': 'open page viewing holders of ERC20 address',
                'action': 'store_true',
            },
            {
                'name': '--logs',
                'help': 'open page viewing logs of transaction',
                'action': 'store_true',
            },
            {
                'name': '--changes',
                'help': 'open page viewing state changes of transaction',
                'action': 'store_true',
            },
            {
                'name': '--abi',
                'help': 'output only the json ABI of an address',
                'action': 'store_true',
            },
        ],
        'help': 'open etherscan query in browser or fetch',
        'examples': {
            '0xd3181ddbb2cea7b4954b8d4a05dbf85d8fc36aef': {'skip': True},
            '0x146063226f2bc60ab02fff825393555672ff505afb352ff11b820355422ba31e': {
                'skip': True
            },
            '14000000': {'skip': True},
            'vitalik.eth': {'skip': True},
            'vitalik.eth --holdings': {'skip': True},
            'vitalik.eth --erc20-transfers': {'skip': True},
            'DAI': {'skip': True},
            'DAI --holders': {'skip': True},
        },
    }


async def async_etherscan_command(
    *,
    query: str,
    logs: bool,
    changes: bool,
    holdings: bool,
    erc20_transfers: bool,
    internal: bool,
    holders: bool,
    abi: bool,
) -> None:

    await etherscan_utils.async_open_etherscan_in_browser(
        query=query,
        logs=logs,
        changes=changes,
        holdings=holdings,
        erc20_transfers=erc20_transfers,
        internal=internal,
        holders=holders,
        abi=abi,
    )




from __future__ import annotations

import typing
from typing_extensions import Literal


# admin schemas = those related to managing ctc
AdminSchemaName = Literal['schema_versions']

# generic schemas = those agnostic to network
GenericSchemaName = Literal[
    '4byte',
    'coingecko',
]

# network schemas = those that have unique data on each network
NetworkSchemaName = Literal[
    # 'block_gas_stats',
    'block_timestamps',
    'block_gas',
    'blocks',
    'contract_abis',
    'contract_creation_blocks',
    'dex_pools',
    'erc20_metadata',
    'events',
    'transactions',
    # 'erc20_state',
    # 'events',
    #
    # protocols
    'chainlink',
]

SchemaName = typing.Union[NetworkSchemaName, GenericSchemaName, AdminSchemaName]




from __future__ import annotations

import typing

import toolstr

import ctc.config
from ctc import cli
from . import llama_requests

if typing.TYPE_CHECKING:
    from typing_extensions import Literal

#
# # pool comparison and aggregation
#


async def async_print_llama_pools_summary(
    *,
    chain: str | None = None,
    project: str | None = None,
    stablecoin: bool | None = None,
    exposure: Literal['single', 'multi'] | None = None,
    min_apy: typing.SupportsFloat | None = None,
    max_apy: typing.SupportsFloat | None = None,
    min_tvl: typing.SupportsFloat | None = None,
    max_tvl: typing.SupportsFloat | None = None,
    n: int | None = None,
    verbose: bool = False,
    show_id: bool = False,
) -> None:

    if n is None:
        n = 15

    pools = await llama_requests.async_get_llama_pools(
        chain=chain,
        project=project,
        stablecoin=stablecoin,
        exposure=exposure,
        min_apy=min_apy,
        max_apy=max_apy,
        min_tvl=min_tvl,
        max_tvl=max_tvl,
    )

    styles = cli.get_cli_styles()

    toolstr.print_text_box('Defi Llama Tracked Pools', style=styles['title'])

    filter_items = [
        chain,
        project,
        stablecoin,
        exposure,
        min_apy,
        max_apy,
        min_tvl,
        max_tvl,
    ]
    if any(item is not None for item in filter_items):
        print()
        toolstr.print('filters:', style=styles['option'])
    if chain is not None:
        toolstr.print(
            '    chain=' + toolstr.add_style(chain, styles['description']),
            style=styles['option'],
        )
    if project is not None:
        toolstr.print(
            '    project=' + toolstr.add_style(project, styles['description']),
            style=styles['option'],
        )
    if stablecoin is not None:
        toolstr.print(
            '    stablecoin='
            + toolstr.add_style(str(stablecoin), styles['description']),
            style=styles['option'],
        )
    if exposure is not None:
        toolstr.print(
            '    exposure='
            + toolstr.add_style(exposure, styles['description']),
            style=styles['option'],
        )
    if min_apy is not None:
        toolstr.print(
            '    min_apy='
            + toolstr.add_style(str(min_apy), styles['description']),
            style=styles['option'],
        )
    if max_apy is not None:
        toolstr.print(
            '    max_apy='
            + toolstr.add_style(str(max_apy), styles['description']),
            style=styles['option'],
        )
    if min_tvl is not None:
        toolstr.print(
            '    min_tvl='
            + toolstr.add_style(str(min_tvl), styles['description']),
            style=styles['option'],
        )
    if max_tvl is not None:
        toolstr.print(
            '    max_tvl='
            + toolstr.add_style(str(max_tvl), styles['description']),
            style=styles['option'],
        )

    print()
    toolstr.print(
        toolstr.add_style(str(len(pools)), styles['description'] + ' bold')
        + ' pools found'
    )
    if verbose:
        print_pool_counts_summary(pools)

    highest_apy_pools = sorted(
        pools, key=lambda pool: float(pool['apy']), reverse=True
    )
    highest_tvl_pools = sorted(
        pools, key=lambda pool: float(pool['tvlUsd']), reverse=True
    )

    print()
    print()
    toolstr.print('Highest APY pools', style=styles['title'])
    print()
    print_pools_table(
        pools=highest_apy_pools,
        n=n,
        show_id=show_id,
    )

    print()
    print()
    toolstr.print('Highest TVL pools', style=styles['title'])
    print()
    print_pools_table(
        pools=highest_tvl_pools,
        n=n,
        show_id=show_id,
    )


def print_pools_table(
    *,
    pools: typing.Sequence[typing.Any],
    n: int,
    show_id: bool = False,
) -> None:

    styles = cli.get_cli_styles()

    labels = [
        'symbol',
        'chain',
        'project',
        'apy',
        'tvl',
    ]
    if show_id:
        labels.append('id')

    rows = []
    for pool in pools[:n]:
        row = [
            pool['symbol'],
            pool['chain'],
            pool['project'],
            pool['apy'],
            pool['tvlUsd'],
        ]
        if show_id:
            row.append(pool['pool'])
        rows.append(row)

    if len(pools) > n:
        rows.append(['...'] * len(labels))

    toolstr.print_table(
        rows,
        labels=labels,
        max_column_widths={'pool': 45},
        column_formats={
            'apy': {'postfix': '%', 'decimals': 2, 'trailing_zeros': True},
            'tvl': {
                'order_of_magnitude': True,
                'prefix': '$',
                'decimals': 2,
                'trailing_zeros': True,
            },
        },
        indent=4,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'symbol': styles['option'],
            'chain': styles['option'],
            'project': styles['option'],
            'apy': styles['description'],
            'tvl': styles['description'],
        },
    )


def print_pool_counts_summary(
    pools: typing.Sequence[typing.Any],
    n: int = 10,
) -> None:

    chain_pool_counts: typing.MutableMapping[str, float | int] = {}
    project_pool_counts: typing.MutableMapping[str, float | int] = {}
    stablecoin_pool_counts: typing.MutableMapping[bool, float | int] = {}

    chain_pool_tvls: typing.MutableMapping[str, float | int] = {}
    project_pool_tvls: typing.MutableMapping[str, float | int] = {}
    stablecoin_pool_tvls: typing.MutableMapping[bool, float | int] = {}

    for pool in pools:

        chain = pool['chain']
        chain_pool_counts.setdefault(chain, 0)
        chain_pool_counts[chain] += 1
        chain_pool_tvls.setdefault(chain, 0)
        chain_pool_tvls[chain] += pool['tvlUsd']

        project = pool['project']
        project_pool_counts.setdefault(project, 0)
        project_pool_counts[project] += 1
        project_pool_tvls.setdefault(project, 0)
        project_pool_tvls[project] += pool['tvlUsd']

        stablecoin = pool['stablecoin']
        stablecoin_pool_counts.setdefault(stablecoin, 0)
        stablecoin_pool_counts[stablecoin] += 1
        stablecoin_pool_tvls.setdefault(stablecoin, 0)
        stablecoin_pool_tvls[stablecoin] += pool['tvlUsd']

    project_pool_counts = dict(
        sorted(
            project_pool_counts.items(), key=lambda item: item[1], reverse=True
        )
    )
    chain_pool_counts = dict(
        sorted(
            chain_pool_counts.items(), key=lambda item: item[1], reverse=True
        )
    )

    money_format: toolstr.NumberFormat = {
        'order_of_magnitude': True,
        'decimals': 2,
        'trailing_zeros': True,
        'prefix': '$',
    }

    styles = cli.get_cli_styles()

    toolstr.print(
        '- stablecoin pools: '
        + toolstr.add_style(
            str(stablecoin_pool_counts[True]), styles['description'] + ' bold'
        ),
    )
    toolstr.print(
        '- stablecoin TVL:',
        toolstr.add_style(
            toolstr.format(stablecoin_pool_tvls[True], **money_format),
            styles['description'] + ' bold',
        ),
    )
    print()
    print()
    toolstr.print('Projects with most pools', style=styles['title'])
    print()
    rows = []
    for project, count in list(project_pool_counts.items())[:n]:
        row = [project, count, project_pool_tvls[project]]
        rows.append(row)
    if len(project_pool_counts) > n:
        rows.append(['...', '...', '...'])
    toolstr.print_table(
        rows,
        labels=['project', 'pools', 'tvl'],
        indent=4,
        column_formats={
            'tvl': money_format,
        },
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'project': styles['option'],
            'pools': styles['description'],
            'tvl': styles['description'],
        },
    )

    print()
    print()
    toolstr.print('Chains with most pools', style=styles['title'])
    print()
    rows = []
    for chain, count in list(chain_pool_counts.items())[:n]:
        row = [chain, count, chain_pool_tvls[chain]]
        rows.append(row)
    if len(chain_pool_counts) > n:
        rows.append(['...', '...', '...'])
    toolstr.print_table(
        rows,
        labels=['chain', 'pools', 'tvl'],
        indent=4,
        column_formats={
            'tvl': money_format,
        },
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'chain': styles['option'],
            'pools': styles['description'],
            'tvl': styles['description'],
        },
    )


#
# # individual pool examination
#


async def async_print_pool_yield_summary(pool: typing.Any) -> None:

    styles = cli.get_cli_styles()

    data = await llama_requests.async_get_pool_yield(pool)

    toolstr.print_text_box(
        'Pool ' + pool + ' History',
        style=styles['title'],
    )
    print()

    plot = toolstr.render_line_plot(
        xvals=data['timestamp'],
        yvals=data['apy'],
        n_rows=10,
        n_columns=60,
        line_style=styles['description'],
        chrome_style=styles['comment'],
        tick_label_style=styles['metavar'],
        yaxis_kwargs={'tick_label_format': {'postfix': '%'}},
        char_dict=ctc.config.get_cli_chart_charset(),
    )
    toolstr.print(
        toolstr.hjustify('APY', 'center', 70),
        indent=4,
        style=styles['title'],
    )
    toolstr.print(plot, indent=4)
    print()

    plot = toolstr.render_line_plot(
        xvals=data['timestamp'],
        yvals=data['tvl'],
        n_rows=10,
        n_columns=60,
        line_style=styles['description'],
        chrome_style=styles['comment'],
        tick_label_style=styles['metavar'],
        yaxis_kwargs={'tick_label_format': {'prefix': '$'}},
        char_dict=ctc.config.get_cli_chart_charset(),
    )
    toolstr.print(
        toolstr.hjustify('TVL', 'center', 70),
        indent=4,
        style=styles['title'],
    )
    toolstr.print(plot, indent=4)



from __future__ import annotations

import typing

import toolsql

from ctc import db
from ctc import spec
from .. import fourbyte_spec


#
# # functions
#


async def async_upsert_function_signature(
    function_signature: fourbyte_spec.PartialEntry,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:
    table = db.get_table_schema('function_signatures', context=None)

    function_signature = function_signature.copy()
    if 'bytes_signature' in function_signature:
        del function_signature['bytes_signature']

    await toolsql.async_insert(
        conn=conn,
        table=table,
        row=function_signature,
        upsert=True,
    )


async def async_upsert_function_signatures(
    function_signatures: typing.Sequence[fourbyte_spec.PartialEntry],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:
    table = db.get_table_schema('function_signatures', context=None)

    function_signatures = []
    for function_signature in list(function_signatures):
        function_signature = function_signature.copy()
        del function_signature['bytes_signature']
        function_signatures.append(function_signature)

    if len(function_signatures) == 0:
        return
    await toolsql.async_insert(
        conn=conn,
        table=table,
        rows=function_signatures,
        upsert=True,
    )


async def async_select_function_signatures(
    conn: toolsql.AsyncConnection,
    *,
    hex_signature: str | None = None,
    text_signature: str | None = None,
    id: int | None = None,
    # bytes_signature: str | None = None,
    context: spec.Context = None,
) -> typing.Sequence[fourbyte_spec.Entry] | None:
    where_equals = {
        'id': id,
        'hex_signature': hex_signature,
        'text_signature': text_signature,
        # 'bytes_signature': bytes_signature,
    }
    where_equals = {
        key: value for key, value in where_equals.items() if value is not None
    }

    table = db.get_table_schema('function_signatures', context=None)

    return await toolsql.async_select(  # type: ignore
        conn=conn,
        table=table,
        where_equals=where_equals,
    )


async def async_delete_function_signatures(
    conn: toolsql.AsyncConnection,
    *,
    hex_signature: str | None = None,
    text_signature: str | None = None,
    context: spec.Context = None,
) -> None:
    where_equals = {
        'hex_signature': hex_signature,
        'text_signature': text_signature,
    }
    where_equals = {
        key: value for key, value in where_equals.items() if value is not None
    }
    if len(where_equals) == 0:
        raise Exception('must specify which feeds to delete')

    table = db.get_table_schema('function_signatures', context=None)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals=where_equals,
    )


#
# # events
#


async def async_upsert_event_signature(
    event_signature: fourbyte_spec.PartialEntry,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:
    table = db.get_table_schema('event_signatures', context=None)

    event_signature = event_signature.copy()
    if 'bytes_signature' in event_signature:
        del event_signature['bytes_signature']

    await toolsql.async_insert(
        conn=conn,
        table=table,
        row=event_signature,
        upsert=True,
    )


async def async_upsert_event_signatures(
    event_signatures: typing.Sequence[fourbyte_spec.PartialEntry],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:
    table = db.get_table_schema('event_signatures', context=None)

    event_signatures = []
    for event_signature in list(event_signatures):
        event_signature = event_signature.copy()
        del event_signature['bytes_signature']
        event_signatures.append(event_signature)

    if len(event_signatures) == 0:
        return
    await toolsql.async_insert(
        conn=conn,
        table=table,
        rows=event_signatures,
        upsert=True,
    )


async def async_select_event_signatures(
    conn: toolsql.AsyncConnection,
    *,
    hex_signature: str | None = None,
    text_signature: str | None = None,
    id: int | None = None,
    # bytes_signature: str | None = None,
    context: spec.Context = None,
) -> typing.Sequence[fourbyte_spec.Entry] | None:
    where_equals = {
        'hex_signature': hex_signature,
        'text_signature': text_signature,
        'id': id,
        # 'bytes_signature': bytes_signature,
    }
    where_equals = {
        key: value for key, value in where_equals.items() if value is not None
    }

    table = db.get_table_schema('event_signatures', context=None)

    return await toolsql.async_select(  # type: ignore
        conn=conn,
        table=table,
        where_equals=where_equals,
    )


async def async_delete_event_signatures(
    conn: toolsql.AsyncConnection,
    *,
    hex_signature: str | None = None,
    text_signature: str | None = None,
    context: spec.Context = None,
) -> None:
    where_equals = {
        'hex_signature': hex_signature,
        'text_signature': text_signature,
    }
    where_equals = {
        key: value for key, value in where_equals.items() if value is not None
    }
    if len(where_equals) == 0:
        raise Exception('must specify which feeds to delete')

    table = db.get_table_schema('event_signatures', context=None)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals=where_equals,
    )




from __future__ import annotations

import typing

import toolcli

from .. import llama_yields

if typing.TYPE_CHECKING:
    from typing_extensions import Literal


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_llama_pools_command,
        'help': 'output data about pools tracked by Defi Llama',
        'args': [
            {
                'name': '--chain',
                'help': 'filter pools by chain',
            },
            {
                'name': '--project',
                'help': 'filter pools by project',
            },
            {
                'name': '--stablecoin',
                'action': 'store_true',
                'help': 'filter pools by whether they are stablecoin-based',
            },
            {
                'name': '--non-stablecoin',
                'action': 'store_true',
                'help': 'filter pools by whether they are not stablecoin-based',
            },
            {
                'name': '--single',
                'action': 'store_true',
                'help': 'filter pools by whether they use a single asset',
            },
            {
                'name': '--multi',
                'action': 'store_true',
                'help': 'filter pools by whether they use a multiple assets',
            },
            {
                'name': '--min-apy',
                'type': float,
                'help': 'filter pools by minimum yield rate',
            },
            {
                'name': '--max-apy',
                'type': float,
                'help': 'filter pools by maximum yield rate',
            },
            {
                'name': '--min-tvl',
                'type': float,
                'help': 'filter pools by miniumum TVL',
            },
            {
                'name': '--max-tvl',
                'type': float,
                'help': 'filter pools by maximum TVL',
            },
            {
                'name': '-n',
                'type': int,
                'help': 'number of pools to display',
            },
            {
                'name': ['-v', '--verbose'],
                'action': 'store_true',
                'help': 'display additional information',
            },
            {
                'name': '--show-id',
                'action': 'store_true',
                'help': 'show id of each pool in table',
            },
        ],
        'examples': [
            '',
            '--min-tvl 1000000',
            '--stablecoin',
            '--chain Ethereum',
            '--min-tvl 1000000 --min-apy 10 --chain Ethereum',
        ],
    }


async def async_llama_pools_command(
    *,
    chain: str | None,
    project: str | None,
    stablecoin: bool,
    non_stablecoin: bool,
    single: bool,
    multi: bool,
    min_apy: typing.SupportsFloat | None = None,
    max_apy: typing.SupportsFloat | None = None,
    min_tvl: typing.SupportsFloat | None = None,
    max_tvl: typing.SupportsFloat | None = None,
    n: int = 10,
    verbose: bool = False,
    show_id: bool = False,
) -> None:

    if single:
        exposure: typing.Optional[Literal['single', 'multi']] = 'single'
    elif multi:
        exposure = 'multi'
    else:
        exposure = None

    if stablecoin:
        use_stablecoin: bool | None = True
    elif non_stablecoin:
        use_stablecoin = False
    else:
        use_stablecoin = None

    await llama_yields.async_print_llama_pools_summary(
        chain=chain,
        project=project,
        stablecoin=use_stablecoin,
        exposure=exposure,
        min_apy=min_apy,
        max_apy=max_apy,
        min_tvl=min_tvl,
        max_tvl=max_tvl,
        n=n,
        verbose=verbose,
        show_id=show_id,
    )



from __future__ import annotations

import typing

from . import typedata

if typing.TYPE_CHECKING:
    from typing_extensions import TypeGuard
    from . import typedefs


def is_schema_name(value: typing.Any) -> TypeGuard[typedefs.SchemaName]:
    return value in typedata.schema_names




from __future__ import annotations

import typing

from ctc import spec

from .. import abi_utils
from .. import event_utils
from . import erc20_metadata
from . import erc20_spec

if typing.TYPE_CHECKING:
    import tooltime


def _get_token_amount_column(df: spec.DataFrame) -> str:
    if 'arg__amount' in df:
        return 'arg__amount'
    elif 'arg__value' in df:
        return 'arg__value'
    elif 'arg__wad' in df:
        return 'arg__wad'
    else:
        raise Exception('could not detect amount column')


async def async_get_erc20_transfers(
    token: spec.ERC20Reference,
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    include_timestamps: bool = False,
    normalize: bool = True,
    verbose: bool = False,
    context: spec.Context = None,
    **event_kwargs: typing.Any,
) -> spec.DataFrame:
    """get transfer events of ERC20 token"""

    token_address = await erc20_metadata.async_get_erc20_address(
        token, context=context
    )

    # use token's own abi if available, otherwise fallback to standard
    # do this because tokens sometimes have different values for event args
    try:
        event_abi = await abi_utils.async_get_event_abi(
            contract_address=token,
            event_name='Transfer',
            context=context,
        )
    except Exception:
        event_abi = erc20_spec.erc20_event_abis['Transfer']

    transfers = await event_utils.async_get_events(
        contract_address=token_address,
        event_abi=event_abi,
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        include_timestamps=include_timestamps,
        verbose=verbose,
        context=context,
        **event_kwargs,
    )

    # make amount column name the same across all tokens
    old_column = 'arg__' + event_abi['inputs'][2]['name']
    column = 'arg__amount'
    transfers = transfers.rename({old_column: column})

    # normalize
    if normalize and len(transfers) > 0:

        import polars as pl
        import numpy as np

        decimals = await erc20_metadata.async_get_erc20_decimals(
            token=token_address,
            block=transfers['block_number'][0],
            context=context,
        )
        factor = float('1e' + str(decimals))
        normalized = np.array(transfers[column].to_list(), dtype=float) / factor
        transfers = transfers.with_columns(pl.Series(column, normalized))

    return transfers


async def async_get_erc20_balances_from_transfers(
    transfers: spec.DataFrame,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    normalize: bool = False,
    context: spec.Context = None,
) -> spec.DataFrame:
    """compute ERC20 balance of each wallet using Transfer events"""

    import polars as pl

    # filter block
    if block is not None:
        transfers = transfers.filter(pl.col('block_number') <= block)

    amount_key = _get_token_amount_column(transfers)

    # subtract transfers out from transfers in
    from_transfers = transfers.groupby('arg__from').agg(pl.sum(amount_key))
    to_transfers = transfers.groupby('arg__to').agg(pl.sum(amount_key))
    balances: spec.DataFrame = to_transfers.sub(from_transfers, fill_value=0)  # type: ignore

    if normalize:
        decimals = await erc20_metadata.async_get_erc20_decimals(
            typing.cast(str, transfers['contract_address'][0]),
            context=context,
        )
        balances /= 10**decimals

    # sort
    balances = balances.sort_values(ascending=False)  # type: ignore

    return balances


def _decode_erc20_transfers(transfers: spec.DataFrame) -> spec.DataFrame:

    import polars as pl
    from ctc.toolbox import pl_utils

    # filter ERC721 based on whether topic3 is indexed
    decoded = transfers.filter(pl.col('topic3').is_null())

    # filter transfers that do not conform to ERC20 specification
    decoded = decoded.filter(
        (~pl.col('topic1').is_null()) & (~pl.col('topic2').is_null())
    )

    if len(decoded) == 0:
        return pl.DataFrame(
            [],
            schema={
                'block_number': pl.Int64,
                'transaction_index': pl.Int64,
                'log_index': pl.Int64,
                'transaction_hash': pl.Binary,
                'contract_address': pl.Binary,
                'from': pl.Binary,
                'to': pl.Binary,
                'value_binary': pl.Binary,
                'value_float': pl.Float64,
            },
        )

    # drop extra columns
    decoded = decoded.drop(['event_hash', 'topic3'])

    # slice topic1 and topic2 to address width
    decoded = decoded.with_columns(
        '0x' + decoded['topic1'].str.slice(-40),
        '0x' + decoded['topic2'].str.slice(-40),
    )

    # strip zeros on unindexed data
    decoded = decoded.with_columns(
        pl.when(pl.col('unindexed').str.lstrip('0x').str.lengths() % 2 == 0)
        .then(pl.col('unindexed').str.lstrip('0x'))
        .otherwise('0' + pl.col('unindexed').str.lstrip('0x'))
    )

    # create value_float column
    decoded = decoded.with_columns(
        decoded['unindexed']
        .apply(lambda x: 0.0 if x == '' else float(int(x, 16)))
        .alias('value_float')
    )

    # convert binary fields
    decoded = pl_utils.prefix_hex_columns_to_binary(
        decoded,
        columns=['transaction_hash', 'contract_address', 'topic1', 'topic2'],
    )
    decoded = pl_utils.raw_hex_columns_to_binary(
        decoded,
        columns=['unindexed'],
    )

    # rename according to ERC20 conventions
    rename = {
        'topic1': 'from',
        'topic2': 'to',
        'unindexed': 'value_binary',
    }
    decoded = decoded.rename(rename)

    return decoded




from __future__ import annotations

import typing

import toolcli
import toolstr

from ctc import cli
from ctc import evm
from ctc import rpc
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_call_command,
        'help': 'output result of a call',
        'args': [
            {'name': 'address', 'help': 'address to point call toward'},
            {
                'name': 'function',
                'help': 'function name, 4byte selector, or ABI',
            },
            {'name': 'args', 'nargs': '*', 'help': 'function arguments'},
            {
                'name': ['--verbose', '-v'],
                'action': 'store_true',
                'help': 'show summary',
            },
            {
                'name': '--from',
                'dest': 'from_address',
                'help': 'address where call should come from',
            },
            {
                'name': '--block',
                'help': 'block number for call',
            },
        ],
        'examples': [
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca totalSupply',
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca 0x18160ddd',
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca \'{"name": "totalSupply", "inputs": [], "outputs": [{"type": "uint256"}]}\'',
        ],
    }


async def async_call_command(
    *,
    address: spec.Address,
    function: str,
    args: typing.Sequence[str],
    verbose: bool,
    from_address: spec.Address,
    block: str,
) -> None:
    import asyncio

    if block is not None:
        block_number = await evm.async_block_number_to_int(block)
    else:
        block_number = 'latest'

    address_task = asyncio.create_task(
        evm.async_resolve_address(address, block=block)
    )

    address = await address_task

    if verbose:
        styles = cli.get_cli_styles()
        toolstr.print_text_box('Performing eth_call', style=styles['title'])
        cli.print_bullet(
            key='to address',
            value=toolstr.add_style(address, styles['metavar']),
        )
        if block is not None:
            cli.print_bullet(key='block', value=block)
        if from_address is not None:
            cli.print_bullet(key='from address', value=from_address)
        cli.print_bullet(key='function', value=function)
        cli.print_bullet(key='arguments', value=args)
        print()
        toolstr.print('result =', style=styles['option'])

    function_abi = await evm.async_parse_function_str_abi(
        function,
        contract_address=address,
    )

    if len(args) != len(function_abi['inputs']):
        raise Exception('improper number of arguments for function')

    # convert args as needed
    typed_args: list[typing.Any] = []
    for arg, arg_abi in zip(args, function_abi['inputs']):
        if arg.isnumeric() and 'int' in arg_abi['type']:
            typed_args.append(int(arg))
        else:
            typed_args.append(arg)

    result = await rpc.async_eth_call(
        to_address=address,
        function_abi=function_abi,
        function_parameters=typed_args,
        from_address=from_address,
        block_number=block_number,
    )
    print(result)



from __future__ import annotations

import typing

from ctc import evm
from ctc import rpc
from ctc import spec

from . import aave_spec


async def async_get_underlying_asset(
    pool_token: spec.Address,
    *,
    context: spec.Context = None,
) -> spec.Address:
    function_abi: spec.FunctionABI = {
        'name': 'UNDERLYING_ASSET_ADDRESS',
        'inputs': [],
        'outputs': [{'type': 'address'}],
    }
    result = await rpc.async_eth_call(
        to_address=pool_token,
        function_abi=function_abi,
        context=context,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_get_reserves_list(
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> typing.Sequence[spec.Address]:

    address = aave_spec.get_aave_address('LendingPool', context=context)

    if block is not None:
        block = evm.standardize_block_number(block)

    reserves: typing.Sequence[spec.Address] = await rpc.async_eth_call(
        to_address=address,
        function_name='getReservesList',
        block_number=block,
        context=context,
    )

    return reserves




from __future__ import annotations

import typing

import toolsql

from ctc import spec
from ... import schema_utils


async def async_upsert_block_transaction_query(
    block_number: int,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:

    table = schema_utils.get_table_schema(
        'block_transaction_queries',
        context=context,
    )
    await toolsql.async_insert(
        conn=conn,
        table=table,
        row={'block_number': block_number},
        upsert=True,
    )


async def async_upsert_block_transaction_queries(
    block_numbers: typing.Sequence[int],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:

    if len(block_numbers) == 0:
        return

    table = schema_utils.get_table_schema(
        'block_transaction_queries',
        context=context,
    )
    rows = [{'block_number': block_number} for block_number in block_numbers]
    await toolsql.async_insert(
        conn=conn,
        table=table,
        rows=rows,
        upsert=True,
    )


async def async_select_block_transaction_queries(
    *,
    block_numbers: typing.Sequence[int] | None = None,
    start_block: int | None = None,
    end_block: int | None = None,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> typing.Sequence[int] | None:

    table = schema_utils.get_table_schema(
        'block_transaction_queries',
        context=context,
    )

    if block_numbers is not None:
        queries: typing.Sequence[int] = await toolsql.async_select(
            conn=conn,
            table=table,
            where_in={'block_number': block_numbers},
            columns=['block_number'],
            output_format='single_column',
        )

    else:
        if start_block is not None:
            where_gte = {'block_number': start_block}
        else:
            where_gte = None
        if end_block is not None:
            where_lte = {'block_number': end_block}
        else:
            where_lte = None
        queries = await toolsql.async_select(
            conn=conn,
            table=table,
            where_lte=where_lte,
            where_gte=where_gte,
            columns=['block_number'],
            output_format='single_column',
        )

    return queries


async def async_select_block_transaction_query(
    block_number: int,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> bool | None:
    """return True if transaction query exists for given block number"""

    table = schema_utils.get_table_schema(
        'block_transaction_queries',
        context=context,
    )

    result = await toolsql.async_select(
        conn=conn,
        table=table,
        where_equals={'block_number': block_number},
        output_format='single_dict_or_none',
    )

    return result is not None


async def async_delete_block_transaction_query(
    block_number: int,
    *,
    context: spec.Context,
    conn: toolsql.AsyncConnection,
) -> None:

    table = schema_utils.get_table_schema(
        'block_transaction_queries', context=context
    )

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals={'block_number': block_number},
    )


async def async_delete_block_transaction_queries(
    block_numbers: typing.Sequence[int],
    *,
    context: spec.Context,
    conn: toolsql.AsyncConnection,
) -> None:

    if len(block_numbers) == 0:
        return

    table = schema_utils.get_table_schema(
        'block_transaction_queries', context=context
    )

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_in={'block_number': block_numbers},
    )




from __future__ import annotations

import typing

from typing_extensions import TypedDict


ChainId = int
NetworkName = str
NetworkReference = typing.Union[ChainId, NetworkName]


class NetworkMetadata(TypedDict):
    name: NetworkName | None
    chain_id: int
    block_explorer: str | None  # url



from __future__ import annotations

import typing

from ctc import evm
from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class FusePool(TypedDict):
        name: str
        creator: spec.Address
        comptroller: spec.Address
        block_posted: int
        timestamp_posted: int

    class FusePoolData(TypedDict):
        total_supply: int
        total_borrow: int
        underlying_tokens: typing.Sequence[spec.Address]
        underlying_symbols: typing.Sequence[str]
        whitelisted_admin: bool

    class FusePoolAsset(TypedDict):
        ftoken: spec.Address
        underlying_token: spec.Address
        underlying_name: str
        underlying_symbol: str
        underlying_decimals: int
        underlying_balance: int
        supply_rate_per_block: int
        borrow_rate_per_block: int
        total_supply: int
        total_borrow: int
        supply_balance: int
        borrow_balance: int
        liquidity: int
        membership: bool
        exchange_rate: int
        underlying_price: int
        oracle: spec.Address
        collateral_factor: int
        reserve_factor: int
        admin_fee: int
        fuse_fee: int
        borrow_guardian_paused: bool

    class FusePoolUser(TypedDict):
        account: spec.Address
        total_borrow: int
        total_collateral: int
        health: int
        assets: typing.Sequence[FusePoolAsset]

    class CTokenOwnership(TypedDict):
        ctoken: spec.Address
        admin: spec.Address
        admin_has_rights: bool
        fuse_admin_has_rights: bool

    class PublicPoolsWithData(TypedDict):
        public_pools: typing.Sequence[FusePool]
        data: typing.Sequence['ReturnPoolSummary']
        errored: typing.Sequence[bool]

    class ReturnPoolSummary(TypedDict):
        total_supply: int
        total_borrow: int
        underlying_tokens: typing.Sequence[spec.Address]
        underlying_symbols: typing.Sequence[str]
        whitelisted_admin: bool

    class ReturnPoolUsersWithData(TypedDict):
        users: typing.Sequence[FusePoolUser]
        close_factor: int
        liquidation_incentive: int

    class ReturnPoolsUsersWithData(TypedDict):
        users: typing.Sequence[typing.Sequence[FusePoolUser]]
        close_factors: typing.Sequence[int]
        liquidation_incentives: typing.Sequence[int]
        errored: typing.Sequence[bool]

    class ReturnPublicPoolUsersWithData(TypedDict):
        comptrollers: typing.Sequence[spec.Address]
        users: typing.Sequence[FusePoolUser]
        close_factors: typing.Sequence[int]
        liquidation_incentives: typing.Sequence[int]
        errored: typing.Sequence[bool]

    class ReturnPoolsBySupplier(TypedDict):
        indices: typing.Sequence[spec.Address]
        account_pools: typing.Sequence[FusePool]

    class ReturnPoolsBySupplierWithData(TypedDict):
        indices: typing.Sequence[spec.Address]
        account_pools: typing.Sequence[FusePool]
        data: typing.Sequence[FusePoolData]
        errored: typing.Sequence[bool]

    class UserSummary(TypedDict):
        supply_balance: int
        borrow_balance: int
        error: bool

    class PoolUserSummary(TypedDict):
        supply_balance: int
        borrow_balance: int

    class ReturnWhitelistedPoolsByAccountWithData(TypedDict):
        indices: typing.Sequence[int]
        account_pools: typing.Sequence[FusePool]
        data: typing.Sequence[FusePoolData]
        errored: typing.Sequence[bool]


def get_lens_address(
    lens_name: str,
    context: spec.Context = None,
) -> spec.Address:
    from ctc import config

    network = config.get_context_chain_id(context)
    network_name = evm.get_network_name(network)

    if network_name == 'ethereum':
        if lens_name == 'primary':
            return '0x6dc585ad66a10214ef0502492b0cc02f0e836eec'
        elif lens_name == 'secondary':
            return '0xc76190e04012f26a364228cfc41690429c44165d'
        else:
            raise Exception('unknown lens name: ' + str(lens_name))

    elif network_name == 'arbitrum':
        if lens_name == 'primary':
            return '0xd6e194af3d9674b62d1b30ec676030c23961275e'
        elif lens_name == 'secondary':
            return '0x32ca4e5d75ecb06f33846055652c831f6e7a6924'
        else:
            raise Exception('unknown lens name: ' + str(lens_name))

    else:
        raise Exception('no lens for network: ' + str(network))


def fuse_pool_to_dict(as_list: typing.Sequence[typing.Any]) -> FusePool:
    keys = list(FusePool.__annotations__.keys())
    if len(as_list) != len(keys):
        raise Exception('invalid number of items')

    return {
        'name': as_list[0],
        'creator': as_list[1],
        'comptroller': as_list[2],
        'block_posted': as_list[3],
        'timestamp_posted': as_list[4],
    }


def fuse_pool_data_to_dict(
    as_list: typing.Sequence[typing.Any],
) -> FusePoolData:
    keys = list(FusePoolData.__annotations__.keys())
    if len(as_list) != len(keys):
        raise Exception('invalid number of items')

    return {
        'total_supply': as_list[0],
        'total_borrow': as_list[1],
        'underlying_tokens': as_list[2],
        'underlying_symbols': as_list[3],
        'whitelisted_admin': as_list[4],
    }


def fuse_pool_asset_to_dict(
    as_list: typing.Sequence[typing.Any],
) -> FusePoolAsset:
    keys = list(FusePoolAsset.__annotations__.keys())
    if len(as_list) != len(keys):
        raise Exception('invalid number of items')

    return {
        'ftoken': as_list[0],
        'underlying_token': as_list[1],
        'underlying_name': as_list[2],
        'underlying_symbol': as_list[3],
        'underlying_decimals': as_list[4],
        'underlying_balance': as_list[5],
        'supply_rate_per_block': as_list[6],
        'borrow_rate_per_block': as_list[7],
        'total_supply': as_list[8],
        'total_borrow': as_list[9],
        'supply_balance': as_list[10],
        'borrow_balance': as_list[11],
        'liquidity': as_list[12],
        'membership': as_list[13],
        'exchange_rate': as_list[14],
        'underlying_price': as_list[15],
        'oracle': as_list[16],
        'collateral_factor': as_list[17],
        'reserve_factor': as_list[18],
        'admin_fee': as_list[19],
        'fuse_fee': as_list[20],
        'borrow_guardian_paused': as_list[21],
    }


def fuse_pool_user_to_dict(
    as_list: typing.Sequence[typing.Any],
) -> FusePoolUser:
    keys = list(FusePoolUser.__annotations__.keys())
    if len(as_list) != len(keys):
        raise Exception('invalid number of items')

    result: FusePoolUser = {
        'account': as_list[0],
        'total_borrow': as_list[1],
        'total_collateral': as_list[2],
        'health': as_list[3],
        'assets': as_list[4],
    }

    result['assets'] = [
        fuse_pool_asset_to_dict(asset)
        for asset in typing.cast(typing.List[typing.Any], result['assets'])
    ]
    return result


def return_pool_summary_to_dict(
    as_list: typing.Sequence[typing.Any],
) -> ReturnPoolSummary:
    keys = list(ReturnPoolSummary.__annotations__.keys())
    if len(as_list) != len(keys):
        raise Exception('invalid number of items')

    return {
        'total_supply': as_list[0],
        'total_borrow': as_list[1],
        'underlying_tokens': as_list[2],
        'underlying_symbols': as_list[3],
        'whitelisted_admin': as_list[4],
    }


def return_pool_users_with_data_to_dict(
    as_list: typing.Sequence[typing.Any],
) -> ReturnPoolUsersWithData:
    keys = list(ReturnPoolUsersWithData.__annotations__.keys())
    if len(as_list) != len(keys):
        raise Exception('invalid number of items')

    result: ReturnPoolUsersWithData = {
        'users': as_list[0],
        'close_factor': as_list[1],
        'liquidation_incentive': as_list[2],
    }

    result['users'] = [
        fuse_pool_user_to_dict(user)
        for user in typing.cast(typing.List[typing.Any], result['users'])
    ]

    return result




from __future__ import annotations

import typing

from ctc import spec

from .. import block_crud
from .. import block_normalize


async def async_get_block_timestamp(
    block: spec.BlockReference,
    *,
    context: spec.Context = None,
) -> int:
    """get timestamp of block"""

    from ctc import config

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='block_timestamps', context=context
    )
    if isinstance(block, int) and read_cache:
        from ctc import db

        timestamp = await db.async_query_block_timestamp(
            block_number=block,
            context=context,
        )
        if timestamp is not None:
            return timestamp

    block_data = await block_crud.async_get_block(block, context=context)
    return block_data['timestamp']


async def async_get_block_timestamps(
    blocks: typing.Sequence[spec.BlockReference],
    *,
    context: spec.Context = None,
) -> list[int]:
    """get timestamps of blocks"""

    from ctc import config

    blocks = await block_normalize.async_block_numbers_to_int(
        blocks=blocks,
        context=context,
    )

    # get timestamps from db
    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='block_timestamps', context=context
    )
    if read_cache:
        from ctc import db

        db_timestamps = await db.async_query_block_timestamps(
            block_numbers=blocks,
            context=context,
        )
        if db_timestamps is None:
            db_timestamps = [None for block in blocks]
        results: dict[int, int | None] = dict(zip(blocks, db_timestamps))
        remaining_blocks: typing.Sequence[int] = [
            block
            for block, timestamp in zip(blocks, db_timestamps)
            if timestamp is None
        ]
    else:
        results = {}
        remaining_blocks = blocks

    # get timestamps from rpc
    if len(remaining_blocks) > 0:
        node_blocks = await block_crud.async_get_blocks(
            blocks=remaining_blocks,
            context=context,
        )
        for block_data in node_blocks:
            results[block_data['number']] = block_data['timestamp']

    output: list[int] = []
    for block in blocks:
        result = results[block]
        if result is None:
            raise Exception('failed to get timestamp for block: ' + str(block))
        output.append(result)
    return output



import pytest

import ctc.config

example_context_config = {
    'default_network': 42161,
    'default_providers': {
        1: 'alchemy__1',
        42161: 'alchemy__42161',
    },
    'providers': {
        'infura__1': {
            'chunk_size': None,
            'convert_reverts_to_none': False,
            'name': 'infura__1',
            'network': 1,
            'protocol': 'http',
            'session_kwargs': {},
            'url': 'http://1.infura.com',
            'disable_batch_requests': False,
        },
        'infura__42161': {
            'chunk_size': None,
            'convert_reverts_to_none': False,
            'name': 'infura__42161',
            'network': 42161,
            'protocol': 'http',
            'session_kwargs': {},
            'url': 'http://42161.infura.com',
            'disable_batch_requests': False,
        },
        'alchemy__1': {
            'chunk_size': None,
            'convert_reverts_to_none': False,
            'name': 'alchemy__1',
            'network': 1,
            'protocol': 'http',
            'session_kwargs': {},
            'url': 'http://1.alchemy.com',
            'disable_batch_requests': False,
        },
        'alchemy__42161': {
            'chunk_size': None,
            'convert_reverts_to_none': False,
            'name': 'alchemy__42161',
            'network': 42161,
            'protocol': 'http',
            'session_kwargs': {},
            'url': 'http://42161.alchemy.com',
            'disable_batch_requests': False,
        },
    },
}

default_network = example_context_config['default_network']
default_provider = example_context_config['providers'][
    example_context_config['default_providers'][default_network]
]

example_context_tests = [
    #
    # defaults
    {
        'context': {},
        'chain_id': default_network,
        'provider_url': default_provider['url'],
    },
    {
        'context': {'provider': None},
        'chain_id': default_network,
        'provider_url': default_provider['url'],
    },
    {
        'context': {
            'provider': None,
        },
        'chain_id': default_network,
        'provider_url': default_provider['url'],
    },
    {
        'context': {'network': None},
        'chain_id': default_network,
        'provider_url': default_provider['url'],
    },
    #
    # provider by name
    {
        'context': {'provider': 'infura__1'},
        'chain_id': 1,
        'provider_url': 'http://1.infura.com',
    },
    {
        'context': {'provider': 'infura__42161'},
        'chain_id': 42161,
        'provider_url': 'http://42161.infura.com',
    },
    {
        'context': {'provider': 'alchemy__1'},
        'chain_id': 1,
        'provider_url': 'http://1.alchemy.com',
    },
    {
        'context': {'provider': 'alchemy__42161'},
        'chain_id': 42161,
        'provider_url': 'http://42161.alchemy.com',
    },
    #
    # provider by known url
    {
        'context': {'provider': 'http://1.infura.com'},
        'chain_id': 1,
        'provider_url': 'http://1.infura.com',
    },
    {
        'context': {'provider': 'http://42161.infura.com'},
        'chain_id': 42161,
        'provider_url': 'http://42161.infura.com',
    },
    {
        'context': {'provider': 'http://1.alchemy.com'},
        'chain_id': 1,
        'provider_url': 'http://1.alchemy.com',
    },
    {
        'context': {'provider': 'http://42161.alchemy.com'},
        'chain_id': 42161,
        'provider_url': 'http://42161.alchemy.com',
    },
    #
    # provider by unknown url
    {
        'context': {'provider': 'https://arbitrumrpc.com'},
        'chain_id': 42161,
        'provider_url': 'https://arbitrumrpc.com',
    },
    #
    # provider by chain id
    {
        'context': {'network': 1},
        'chain_id': 1,
        'provider_url': 'http://1.alchemy.com',
    },
    {
        'context': {'network': 42161},
        'chain_id': 42161,
        'provider_url': 'http://42161.alchemy.com',
    },
    #
    # provider by network name
    {
        'context': {'network': 'ethereum'},
        'chain_id': 1,
        'provider_url': 'http://1.alchemy.com',
    },
    {
        'context': {'network': 'arbitrum'},
        'chain_id': 42161,
        'provider_url': 'http://42161.alchemy.com',
    },
]


@pytest.fixture()
def set_example_context_config():
    import ctc.config

    for key, value in example_context_config.items():
        ctc.config.set_config_override(key=key, value=value)

    yield

    ctc.config.clear_all_config_overrides()


@pytest.mark.parametrize('test', example_context_tests)
def test_get_context_chain_id(test, set_example_context_config):
    context = test['context']
    target_chain_id = test['chain_id']
    actual_chain_id = ctc.config.get_context_chain_id(context)
    assert actual_chain_id == target_chain_id


@pytest.mark.parametrize('test', example_context_tests)
def test_get_context_provider(test, set_example_context_config):
    context = test['context']
    if 'provider_url' in test:
        target_provider_url = test['provider_url']
        provider = ctc.config.get_context_provider(context)
        actual_provider_url = provider['url']
        assert actual_provider_url == target_provider_url




from __future__ import annotations

import typing

from ctc import spec
from .. import rpc_logging


def create(
    method: str, parameters: list[typing.Any]
) -> spec.RpcSingularRequest:
    """create single RPC request given the RPC method and parameters"""

    import random

    return {
        'jsonrpc': '2.0',
        'method': method,
        'params': parameters,
        'id': random.randint(1, int(1e18)),
    }


def _postprocess_response(
    *,
    raw_response: str,
    request: spec.RpcRequest,
    provider: spec.Provider,
    raw_output: bool,
    convert_reverts_to: typing.Any = None,
    convert_reverts_to_none: bool = False,
    logging_rpc_calls: bool,
) -> typing.Any:

    if raw_output:
        return raw_response
    else:
        import orjson

        response = orjson.loads(raw_response)

    if 'result' not in response and 'error' in response:
        if provider['convert_reverts_to_none'] or convert_reverts_to_none:
            output = None
        elif convert_reverts_to is not None:
            output = convert_reverts_to
        else:
            if typing.TYPE_CHECKING:
                response = typing.cast(
                    spec.RpcSingularResponseFailure, response
                )
            raise spec.RpcException(
                'RPC ERROR: ' + response['error']['message']
            )
    else:
        if typing.TYPE_CHECKING:
            response = typing.cast(
                spec.RpcSingularResponseSuccess, response
            )
        output = response['result']

    if logging_rpc_calls:
        rpc_logging.log_rpc_response(
            response=response,
            request=request,
            provider=provider,
        )

    return output


def _postprocess_plural_response(
    *,
    raw_response_chunks: typing.Sequence[str],
    request_chunks: list[spec.RpcPluralRequest],
    request: spec.RpcRequest,
    provider: spec.Provider,
    raw_output: bool,
    convert_reverts_to: typing.Any = None,
    convert_reverts_to_none: bool = False,
    logging_rpc_calls: bool,
) -> str | typing.Any:

    if raw_output:
        return raw_response_chunks
    else:
        import orjson

        response_chunks = [
            orjson.loads(raw_response_chunk)
            for raw_response_chunk in raw_response_chunks
        ]

    if logging_rpc_calls:
        for request_chunk, response_chunk in zip(
            request_chunks, response_chunks
        ):
            rpc_logging.log_rpc_response(
                response=response_chunk,
                request=request_chunk,
                provider=provider,
            )

    # reorder chunks
    assert isinstance(request, list)
    plural_response = _reorder_response_chunks(response_chunks, request)

    if (
        provider['convert_reverts_to_none']
        or convert_reverts_to_none
        or convert_reverts_to is not None
    ):
        if convert_reverts_to is not None:
            revert_value = convert_reverts_to
        else:
            revert_value = None
        output = []
        for subresponse in plural_response:
            if 'result' in subresponse:
                output.append(subresponse['result'])
            elif 'error' in subresponse:
                output.append(revert_value)
            else:
                raise Exception('could not process response')
    else:
        output = [subresponse['result'] for subresponse in plural_response]

    return output


def _reorder_response_chunks(
    response_chunks: list[spec.RpcPluralResponseRaw],
    request: spec.RpcPluralRequest,
) -> spec.RpcPluralResponse:
    responses_by_id = {
        response['id']: response
        for response_chunk in response_chunks
        for response in response_chunk
    }
    return [responses_by_id[subrequest['id']] for subrequest in request]


#
# # chunking
#


def _chunk_request(
    request: spec.RpcPluralRequest, provider: spec.Provider
) -> list[spec.RpcPluralRequest]:
    if provider['chunk_size'] is not None:
        return _chunk_request_by_size(request, provider['chunk_size'])
    else:
        return [request]


def _chunk_request_by_size(
    request: spec.RpcPluralRequest, chunk_size: int
) -> list[spec.RpcPluralRequest]:
    import math

    n_chunks = math.ceil(len(request) / chunk_size)
    return [
        request[slice(c * chunk_size, (c + 1) * chunk_size)]
        for c in range(n_chunks)
    ]


def _chunk_request_by_method(request: spec.RpcRequest) -> None:
    raise NotImplementedError()


def _chunk_request_by_block_range(request: spec.RpcRequest) -> None:
    raise NotImplementedError()




from __future__ import annotations

import functools
from typing import Callable, Coroutine, Any, TypeVar

import toolsql

from ctc import spec
from . import connect_utils


R = TypeVar('R')


def wrap_selector_with_connection(
    async_f: Callable[..., Coroutine[Any, Any, R | None]],
    schema_name: spec.SchemaName | Callable[..., spec.SchemaName | None],
) -> Callable[..., Coroutine[Any, Any, R | None]]:

    # define new function
    @functools.wraps(async_f)
    async def async_connected_f(
        *args: Any,
        context: spec.Context,
        db_config: toolsql.DBConfig | None = None,
        raise_if_cannot_connect: bool = False,
        raise_if_table_dne: bool = False,
        **kwargs: Any,
    ) -> R | None:

        # determine schema_name
        if not isinstance(schema_name, str) and hasattr(
            schema_name, '__call__'
        ):
            name = schema_name()
            if name is None:
                return None
        elif isinstance(schema_name, str):
            name = schema_name
        else:
            raise Exception('unknown schema_name format')

        # try to query data
        try:
            async with connect_utils.async_connect(
                context=context,
                db_config=db_config,
                schema=name,
                read_only=True,
            ) as conn:
                return await async_f(*args, conn=conn, context=context, **kwargs)

        # handle errors, either by returning None or raising exceptions
        except toolsql.CannotConnect as e:
            if raise_if_cannot_connect:
                raise e
            else:
                return None
        except toolsql.TableDoesNotExist as e:
            if raise_if_table_dne:
                raise e
            else:
                return None

    # return decorated function
    return async_connected_f




from __future__ import annotations

import typing

from ctc import spec
from ctc import evm

from .. import chainlink_spec
from . import feed_datum


async def async_get_feed_answer_datum_by_block(
    feed: chainlink_spec._FeedReference,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    context: spec.Context = None,
    normalize: bool = True,
    interpolate: bool = True,
    invert: bool = False,
) -> spec.DataFrame:
    import asyncio
    from ctc.toolbox import pl_utils
    import polars as pl

    int_blocks = await evm.async_block_numbers_to_int(
        blocks=blocks, context=context
    )

    # query data
    coroutines = []
    for block in int_blocks:
        coroutine = feed_datum.async_get_feed_datum(
            feed=feed,
            fields='answer',
            normalize=normalize,
            block=block,
            invert=invert,
            context=context,
        )
        coroutines.append(coroutine)
    result = await asyncio.gather(*coroutines)

    # create series
    df = pl.DataFrame({'block_number': int_blocks, 'answer': result})

    # interpolate blocks
    if interpolate:
        df = pl_utils.interpolate(df, index_column='block_number')

    return df


async def async_get_feed_full_datum_by_block(
    feed: chainlink_spec._FeedReference,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    context: spec.Context = None,
    normalize: bool = True,
    interpolate: bool = True,
    invert: bool = False,
) -> spec.DataFrame:
    import asyncio
    import polars as pl
    from ctc.toolbox import pl_utils

    int_blocks = await evm.async_block_numbers_to_int(
        blocks=blocks, context=context
    )

    # query data
    coroutines = []
    for block in int_blocks:
        coroutine = feed_datum.async_get_feed_datum(
            feed=feed,
            block=block,
            normalize=normalize,
            fields='full',
            invert=invert,
            context=context,
        )
        coroutines.append(coroutine)
    result = await asyncio.gather(*coroutines)

    # create series
    df = pl.DataFrame({'block_number': int_blocks, 'answer': result})

    # interpolate blocks
    if interpolate:
        df = pl_utils.interpolate(df, index_column='block_number')

    return df




from __future__ import annotations

import toolsql


transactions_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'transactions': {
            'columns': [
                {'name': 'hash', 'type': 'Text', 'primary': True},
                {'name': 'block_number', 'type': 'Integer', 'index': True},
                {'name': 'transaction_index', 'type': 'Integer'},
                {'name': 'to_address', 'type': 'Text', 'index': True},
                {'name': 'from_address', 'type': 'Text', 'index': True},
                {'name': 'value', 'type': 'Text'},  # int <-> str
                {'name': 'input', 'type': 'Text'},
                {'name': 'nonce', 'type': 'Integer'},
                {'name': 'transaction_type', 'type': 'Integer'},
                {'name': 'status', 'type': 'Boolean'},  # from receipt
                {'name': 'gas_used', 'type': 'Integer'},  # from receipt
                {'name': 'gas_limit', 'type': 'Integer'},
                {'name': 'gas_priority', 'type': 'Integer', 'nullable': True},  # eip1559
                {'name': 'gas_price', 'type': 'Integer'},
                {'name': 'gas_price_max', 'type': 'Integer', 'nullable': True},  # eip1559
                {'name': 'access_list', 'type': 'JSON', 'nullable': True},  # type 1 and eip1559 tx
            ],
        },
        'block_transaction_queries': {
            'columns': [
                {'name': 'block_number', 'type': 'Integer', 'primary': True},
            ],
        },
    },
}




from __future__ import annotations

import toolsql


erc20_state_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'erc20_balances': {
            'columns': [
                {
                    'name': 'erc20',
                    'type': 'Text',
                    'primary': True,
                },
                {
                    'name': 'wallet',
                    'type': 'Text',
                    'primary': True,
                },
                {
                    'name': 'block',
                    'type': 'Integer',
                    'primary': True,
                },
                {
                    'name': 'balance',
                    'type': 'Text',
                },
            ],
        },
        'erc20_total_supplies': {
            'columns': [
                {
                    'name': 'erc20',
                    'type': 'Text',
                    'index': True,
                },
                {
                    'name': 'block',
                    'type': 'Integer',
                    'index': True,
                },
                {
                    'name': 'total_supply',
                    'type': 'Text',
                },
            ],
        },
    },
}



import pytest

from ctc import evm
from ctc.evm.erc20_utils import erc20_spec


DAI_CONTRACT = "0x6B175474E89094C44Da98b954EedeAC495271d0F".lower()
START_BLOCK = 15181140
START_TIMESTAMP = 1658342398
END_BLOCK = 15181147


@pytest.mark.asyncio
async def test_get_events():
    df = await evm.async_get_events(
        contract_address=DAI_CONTRACT,
        event_abi=erc20_spec.erc20_event_abis['Transfer'],
        start_block=START_BLOCK,
        end_block=END_BLOCK,
        include_timestamps=True,
        event_name="Transfer",
    )
    assert len(df) == 6
    assert df["contract_address"][0] == DAI_CONTRACT
    assert df["block_number"][0] == START_BLOCK
    assert df["timestamp"][0] == START_TIMESTAMP



from __future__ import annotations

from ctc import spec


def _validate_context(context: spec.Context) -> None:
    """validate user-given context specification"""

    if isinstance(context, dict):
        if not set(context.keys()).issubset(spec.context_keys):
            raise Exception(
                'invalid keys for context: '
                + str(set(context.keys()) - set(spec.context_keys))
                + ', valid keys are: '
                + str(set(spec.context_keys))
            )




"""the goal of these functions is to return the state of the ctc configuration

these functions should
- allow user data to be somewhat malformed
    - assume types are proper
    - some values might be missing
    - str values might be invalid
- return a default when appropriate if no config value is specified
"""

from __future__ import annotations

import typing
import os

if typing.TYPE_CHECKING:
    import toolcli
    import toolsql
    import toolstr

from ctc import spec
from . import config_defaults
from . import config_read


def get_config_spec_version() -> str | None:
    return config_read.get_config().get('config_spec_version')


def get_data_dir() -> str:
    data_dir = config_read.get_config().get('data_dir')
    if data_dir is not None:
        data_dir = os.path.expanduser(data_dir)
        return data_dir
    else:
        return config_defaults.get_default_data_dir()


#
# # networks
#


def get_default_network() -> spec.ChainId:
    return config_read.get_config()['default_network']


def get_config_networks() -> typing.Mapping[spec.ChainId, spec.NetworkMetadata]:
    config = config_read.get_config()
    networks = config.get('networks')
    if networks is not None:
        return networks
    else:
        return config_defaults.get_default_networks_metadata()


def get_networks_that_have_providers() -> typing.Sequence[spec.ChainId]:
    providers = get_providers()
    all_networks = get_config_networks()
    network_set: set[spec.ChainId] = set()
    for provider in providers.values():
        network = provider.get('network')
        if network is not None:

            # convert chain_id to name
            if isinstance(network, str):
                for chain_id, network_metadata in all_networks.items():
                    if network == network_metadata.get('name'):
                        network_set.add(chain_id)
                        break
                else:
                    raise Exception('unknown network name: ' + str(network))

            # add network
            if isinstance(network, int):
                network_set.add(network)
            else:
                raise Exception('unknown network specification')

    return sorted(network for network in network_set)


#
# # providers
#


def get_providers() -> typing.Mapping[spec.NetworkName, spec.Provider]:
    return config_read.get_config()['providers']


def get_network_default_provider(
    network: spec.NetworkName | spec.ChainId,
) -> spec.Provider | None:
    """get default provider for network"""

    if not isinstance(network, int):
        if isinstance(network, str):
            for chain_id, network_metadata in get_config_networks().items():
                if network == network_metadata['name']:
                    network = chain_id
                    break
            else:
                raise Exception('unknown network name: ' + str(type(network)))
        else:
            raise Exception('unknown network type: ' + str(type(network)))

    # get provider of network
    config = config_read.get_config()
    default_providers = config.get('default_providers', {})
    if network in default_providers:
        provider_name = default_providers[network]
        return config['providers'][provider_name]
    else:
        message = 'no default provider specified for network ' + str(network)
        raise Exception(message)


#
# # caches
#

def get_context_cache_rules() -> typing.Sequence[spec.ContextCacheRule]:
    config = config_read.get_config()
    return config['context_cache_rules']


#
# # db
#

def get_db_config(backend_name: str) -> toolsql.DBConfig:
    config = config_read.get_config()
    db_config = config['db_configs'].get(backend_name)
    if db_config is None:
        raise Exception('no valid db_config for ' + str(backend_name))
    return db_config


def get_cache_backends() -> typing.Sequence[str]:
    config = config_read.get_config()
    return list(config['db_configs'].keys())


#
# # logging
#


def get_log_rpc_calls() -> bool:
    config = config_read.get_config()
    return config.get('log_rpc_calls', False)


def get_log_sql_queries() -> bool:
    config = config_read.get_config()
    return config.get('log_sql_queries', False)


def get_log_dir() -> str:
    return os.path.join(get_data_dir(), 'logs')


def get_rpc_requests_log_path() -> str:
    log_dir = get_log_dir()
    os.makedirs(log_dir, exist_ok=True)
    return os.path.join(log_dir, 'rpc_requests.log')


def get_sql_queries_log_path() -> str:
    log_dir = get_log_dir()
    os.makedirs(log_dir, exist_ok=True)
    return os.path.join(log_dir, 'sql_queries.log')


#
# # cli
#


def get_cli_color_theme() -> toolcli.StyleTheme:
    config = config_read.get_config()
    return config['cli_color_theme']


def get_cli_chart_charset() -> toolstr.SampleMode:
    config = config_read.get_config()
    return config['cli_chart_charset']




from __future__ import annotations

from . import uniswap_v2_dex


class SushiDEX(uniswap_v2_dex.UniswapV2DEX):
    """Sushi DEX"""

    _pool_factories = {1: ['0xc0aee478e3658e2610c5f7a4a2e1777ce9e4f2ac']}



from __future__ import annotations

import typing

from ctc import spec
from .. import binary_utils
from . import transaction_types


def serialize_unsigned_transaction(
    transaction: spec.PrechainTransaction,
    *,
    chain_id: int | None = None,
) -> spec.PrefixHexData:
    """serialize unsigned transaction"""

    # collect keys for transaction type
    type = transaction_types.get_transaction_type(transaction)
    keys = transaction_types.get_transaction_type_keys(transaction, signed=False)

    # represent as list
    as_list: list[typing.Any] = []
    for key in keys:
        if key == 'chain_id':
            if transaction.get('chain_id') is not None:
                as_list.append(transaction.get('chain_id'))
            elif transaction.get('chainId') is not None:
                as_list.append(transaction.get('chainId'))
            elif chain_id is not None:
                as_list.append(chain_id)
            else:
                raise Exception('must specify chain_id')
        elif key in ['access_list', 'accessList']:
            access_list = transaction[key]  # type: ignore
            if len(access_list) > 0 and isinstance(access_list[0], dict):
                if 'storageKeys' in access_list[0]:
                    access_list = [
                        [item['address'], item['storageKeys']]
                        for item in access_list
                    ]
                else:
                    access_list = [
                        [item['address'], item['storage_keys']]
                        for item in access_list
                    ]
            as_list.append(access_list)
        else:
            as_list.append(transaction[key])  # type: ignore

    # add EIP-155 chain_id
    if type == 0:
        if chain_id is None and transaction.get('chain_id') is not None:
            chain_id = binary_utils.binary_convert(transaction.get('chain_id'), 'integer')  # type: ignore
        if chain_id is not None:
            as_list.append(chain_id)
            as_list.append(0)
            as_list.append(0)

    # get transaction type prefix
    if type == 0:
        prefix = '0x'
    elif type == 1:
        prefix = '0x01'
    elif type == 2:
        prefix = '0x02'
    else:
        raise Exception('unknown transaction type: ' + str(type))

    return prefix + binary_utils.rlp_encode(as_list, 'raw_hex')


def serialize_signed_transaction(
    transaction: spec.PrechainTransaction,
) -> spec.PrefixHexData:
    """serialize signed transaction"""

    # collect keys for transaction type
    type = transaction_types.get_transaction_type(transaction)
    keys = transaction_types.get_transaction_type_keys(transaction, signed=True)

    # get list of fields
    as_list = [transaction[key] for key in keys]  # type: ignore

    # get transaction type prefix
    if type == 0:
        prefix = '0x'
    elif type == 1:
        prefix = '0x01'
    elif type == 2:
        prefix = '0x02'
    else:
        raise Exception('unknown transaction type: ' + str(type))

    # encode as rlp
    return prefix + binary_utils.rlp_encode(as_list, 'raw_hex')




import pytest

from ctc import evm
from ctc import rpc


example_txs = [
    '0xccb0a942a36db42ccc5ee226e4f0599c761d4c0a884f3ac2f7a56fef7aef85df',
    '0xfd74b4443e147687326e76886417a5a81c07ba26568c87e928546ef5b8dacd0d',
    '0x0bf99b11e4f4963d34b5b9d24a542ff045d8e436740dcda698ec29deff84c959',
]


@pytest.mark.parametrize('test', example_txs)
async def test_get_tx_hash(test):

    target_hash = test
    transaction = await rpc.async_eth_get_transaction_by_hash(target_hash)
    actual_hash = evm.hash_signed_transaction(transaction)
    assert actual_hash == target_hash


@pytest.mark.parametrize('test', example_txs)
async def test_verify_transaction_signature(test):

    tx_hash = test
    transaction = await rpc.async_eth_get_transaction_by_hash(tx_hash)
    signature = (
        transaction['v'],
        transaction['r'],
        transaction['s'],
    )
    assert evm.verify_transaction_signature(
        signature=signature,
        transaction=transaction,
        address=transaction['from'],
    )


example_transactions = [
    {
        # from https://github.com/ethereum/web3.py/blob/master/docs/web3.eth.account.rst#sign-a-transaction
        'private_key': '0x4c0883a69102937d6231471b5dbb6204fe5129617082792ae468d01a3f362318',
        'signature': (
            38,
            84095564551732371065849105252408326384410939276686534847013731510862163857293,
            32698347985257114675470251181312399332782188326270244072370350491677872459742,
        ),
        'transaction': {
            'to': '0xF0109fC8DF283027b6285cc889F5aA624EaC1F55',
            'value': 1000000000,
            'gas': 2000000,
            'max_fee_per_gas': 2000000000,
            'max_priority_fee_per_gas': 1000000000,
            'nonce': 0,
            'chain_id': 1,
            'type': 2,
            'access_list': (
                [
                    '0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae',
                    [
                        '0x0000000000000000000000000000000000000000000000000000000000000003',
                        '0x0000000000000000000000000000000000000000000000000000000000000007',
                    ],
                ],
                [
                    '0xbb9bc244d798123fde783fcc1c72d3bb8c189413',
                    [],
                ],
            ),
            'input': bytes(),
        },
    },
    {
        # from https://github.com/ethereum/web3.py/blob/master/docs/web3.eth.account.rst#sign-a-transaction
        'private_key': b"\xb2\\}\xb3\x1f\xee\xd9\x12''\xbf\t9\xdcv\x9a\x96VK-\xe4\xc4rm\x03[6\xec\xf1\xe5\xb3d",
        'signature': (
            38,
            93522894155654168208483453926995743737629589441154283159505514235904280342434,
            48417310681110102814014302147799665717176259465062324746227758019974374282313,
        ),
        'transaction': {
            'value': 0,
            'chainId': 1,
            'gas': 70000,
            'max_fee_per_gas': 2000000000,
            'max_priority_fee_per_gas': 1000000000,
            'nonce': 0,
            'to': '0xfB6916095ca1df60bB79Ce92cE3Ea74c37c5d359',
            'input': '0xa9059cbb000000000000000000000000fb6916095ca1df60bb79ce92ce3ea74c37c5d3590000000000000000000000000000000000000000000000000000000000000001',
            'access_list': [],
        },
    },
]


@pytest.mark.parametrize('test', example_transactions)
def test_sign_transaction(test):

    private_key = test['private_key']
    transaction = test['transaction']
    target_signature = test['signature']

    # transaction = evm.standardize_transaction(transaction)

    actual_signature = evm.sign_transaction(
        transaction=transaction,
        private_key=private_key,
    )

    assert actual_signature == target_signature



from __future__ import annotations

import typing

import toolsql

from ctc import evm
from ctc import spec
from ... import schema_utils


async def async_upsert_transaction(
    db_transaction: spec.DBTransaction,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:

    table = schema_utils.get_table_schema('transactions', context=context)

    tx = evm.convert_db_transaction_fields_to_text(db_transaction)

    await toolsql.async_insert(
        conn=conn,
        table=table,
        row=tx,
        upsert=True,
    )


async def async_upsert_transactions(
    transactions: typing.Sequence[spec.DBTransaction],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:

    table = schema_utils.get_table_schema('transactions', context=context)

    if len(transactions) == 0:
        return

    txs = [evm.convert_db_transaction_fields_to_text(tx) for tx in transactions]

    await toolsql.async_insert(
        conn=conn,
        table=table,
        rows=txs,
        upsert=True,
    )


async def async_select_transaction(
    hash: str,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> spec.DBTransaction | None:

    table = schema_utils.get_table_schema('transactions', context=context)

    columns = [
        column['name']
        for column in table['columns']
        if column['name'] != 'access_list'
    ]

    tx: spec.DBTransactionText | None = await toolsql.async_select(  # type: ignore
        conn=conn,
        table=table,
        columns=columns,
        where_equals={'hash': hash},
        output_format='single_dict_or_none',
    )

    if tx is None:
        return None
    else:
        return evm.convert_db_transaction_fields_to_int(tx)


async def async_select_transactions(
    hashes: typing.Sequence[str],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> typing.Sequence[spec.DBTransaction | None] | None:

    table = schema_utils.get_table_schema('transactions', context=context)

    columns = [
        column['name']
        for column in table['columns']
        if column['name'] != 'access_list'
    ]

    transactions: typing.Sequence[
        spec.DBTransaction
    ] = await toolsql.async_select(  # type: ignore
        conn=conn,
        table=table,
        columns=columns,
        where_in={'hash': hashes},
    )

    if transactions is None:
        return None

    result = {tx['hash']: tx for tx in transactions}

    return [result.get(tx_hash) for tx_hash in hashes]


async def async_delete_transaction(
    hash: str,
    *,
    context: spec.Context,
    conn: toolsql.AsyncConnection,
) -> None:

    table = schema_utils.get_table_schema('transactions', context=context)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals={'hash': hash},
    )


async def async_delete_transactions(
    hashes: typing.Sequence[str],
    *,
    context: spec.Context,
    conn: toolsql.AsyncConnection,
) -> None:

    if len(hashes) == 0:
        return

    table = schema_utils.get_table_schema('transactions', context=context)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_in={'hash': hashes},
    )




from __future__ import annotations

import typing

from ctc import evm
from ctc import rpc
from ctc import spec

from . import contracts
from . import uniswap_v3_spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    import tooltime

    class UniswapV3PoolMetadata(TypedDict):
        x_symbol: str
        y_symbol: str
        x_address: str
        y_address: str
        fee: int


#
# # metadata
#


async def async_get_pool_tokens(
    pool_address: spec.Address,
    *,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> tuple[spec.Address, spec.Address]:
    import asyncio

    token0_abi = await uniswap_v3_spec.async_get_function_abi('token0', 'pool')
    token1_abi = await uniswap_v3_spec.async_get_function_abi('token1', 'pool')
    kwargs = dict(rpc_kwargs, to_address=pool_address, context=context)
    return await asyncio.gather(
        rpc.async_eth_call(function_abi=token0_abi, **kwargs),
        rpc.async_eth_call(function_abi=token1_abi, **kwargs),
    )


async def async_get_pool_metadata(
    pool_address: spec.Address,
    *,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> UniswapV3PoolMetadata:
    x_address, y_address = await async_get_pool_tokens(
        pool_address=pool_address, context=context
    )
    x_symbol, y_symbol = await evm.async_get_erc20s_symbols(
        tokens=[x_address, y_address], context=context, **rpc_kwargs
    )
    fee = await contracts.async_pool_fee(pool_address, context=context)
    return {
        'x_symbol': x_symbol,
        'y_symbol': y_symbol,
        'x_address': x_address,
        'y_address': y_address,
        'fee': fee,
    }


#
# # events
#


async def async_get_pool_swaps(
    pool_address: spec.Address,
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    include_timestamps: bool = False,
    replace_symbols: bool = False,
    normalize: bool = True,
    context: spec.Context = None,
) -> spec.DataFrame:
    import asyncio
    import polars as pl

    if normalize or replace_symbols:
        metadata_task = asyncio.create_task(
            async_get_pool_metadata(pool_address, context=context)
        )

    event_abi = await uniswap_v3_spec.async_get_event_abi('Swap', 'pool')

    swaps = await evm.async_get_events(
        event_abi=event_abi,
        contract_address=pool_address,
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        include_timestamps=include_timestamps,
        context=context,
    )

    if normalize or replace_symbols:
        metadata = await metadata_task

    # rename columns
    if replace_symbols:
        x_symbol = metadata['x_symbol']
        y_symbol = metadata['y_symbol']
    else:
        x_symbol = 'x'
        y_symbol = 'y'
    columns = {
        'arg__amount0': x_symbol + '_amount',
        'arg__amount1': y_symbol + '_amount',
    }
    swaps = swaps.rename(columns)

    # normalize columns
    if normalize:
        x_decimals, y_decimals = await evm.async_get_erc20s_decimals(
            tokens=[metadata['x_address'], metadata['y_address']],
            context=context,
        )
        swaps = swaps.with_columns(
            pl.col(columns['arg__amount0']).cast(float) / (10**x_decimals),
            pl.col(columns['arg__amount1']).cast(float) / (10**y_decimals),
        )

    return swaps

